import time
start_time = time.time()
import math
import numpy as np
from tqdm import tqdm
from data_preprocessing import one_hot, revert_one_hot
import pandas as pd
data = pd.read_csv('/users/anish/downloads/breast-cancer-wisconsin.data')

def data_collection(data) :
    training_data = data[['ID', 'clump_thickness','cell_uniformity', 'shape_uniformity', 'chromatin','nucleoli']]
    labels = data['class']

    training_data = training_data.drop(['ID', 'chromatin'], axis = 1)
    from sklearn.preprocessing import StandardScaler
    scaler = StandardScaler()
    training_data = training_data._get_numeric_data()

    training_data = scaler.fit_transform(training_data)

    training_data, labels = np.array(training_data), np.array(labels)

    split = 30
    x_test, y_test = training_data[:split], labels[:split]
    x_train,y_train = training_data[split:], labels[split:]

    x_train, y_train, x_test, y_test = np.array(x_train, dtype = float), np.array(y_train, dtype = float), np.array(x_test, dtype = float), np.array(y_test, dtype = float)
    y_train, y_test = y_train/2 - 1, y_test/2 - 1

    return x_train,x_test, y_train, y_test

x_train, x_test, y_train, y_test = data_collection(data)


def _preprocess_data(data, labels):
    """
    Convert all data + labels to numpy arrays, and make data 2D if needed and labels 2D (for it to work with subtraction with the np.dot, etc.)
    :param self:
    """
    data, labels = np.array(data), np.array(labels)

    if len(data.shape) < 2:
        data = np.array([np.array([x_i]) for x_i in data])

    if len(labels.shape) < 2:
        labels = np.array([np.array([y_i]) for y_i in labels])

    return data, labels


class LinearRegression():
    def __init__(self):
        self.information = """

                Linear Regression is arguably the most basic and simple machine learning algorithm there is. There is a huge chance that you have used this algorithm in your math class in the past or heard of it. All it tries to
                do is find the line of best fit in data. Seems simple to do with paper, but how about in code? And what if the graph is 100+ dimensions? How would you find the best line in the first place? All of this creates 
                an optimization problem for finding the line that reduces the residuals (distance from point and line) to as much as possible. Linear Regression utilizes an algorithm known as gradient descent, also found in 
                all other regression models and heavily in neural networks today. Make sure to read on, as the lines below explain how to use the library (the syntax is consistent for all models.)

                To train a model : - Data must be 2D, meaning a list/vector of list/vectors. For example : [[1, 1, 1], [2, 2, 2], [3, 3, 3]] - All data points must be of the same size. This works : [[1, 1], [1, 2], [0, 1]]. 
                This doesn't : [[0], [1, 2], [3, 4]] (first data point has only a size of 1) - labels must 
                be stored in a 1D dataset like : [1, 2, 3, 4, 4]. You cannot do : [[1, 2, 3, 4]] or [[1], [2], [3]]. 

                Onto code! First step is to define a model to use its functions.

                model = LinearRegression()

                Training the model : 

                model.fit(data, labels)  # note : certain models will have certain keywords required. Call model.information to see that

                Evaluating model (you want a higher number here): 

                model.evaluate(data, labels) #data must be 2D, labels must be 1D
                ex. data = [[1, 1]]; labels = [1]
                ex. data = [[1, 1], [2, 2], [3, 3]], labels = [1, 2, 3]

                Predict : 

                model.predict(data) #data must be 2D, even if just one sample (ex. model.predict([[1, 2]])!

                © Anish Lakkapragada 2020   
        """

    def reset_params(self):
        '''when model predictions aren't good you can always do model.reset_params() and retrain. Always train before resetting parameters.'''
        self.weights = np.random.randn(self.num_features, 1)
        self.bias = np.random.uniform()

    def fit(self, x_train, y_train, accuracy_desired=0.95, learning_rate=0.01, max_iters=10000, show_acc=True):
        """
        :param x_train: input (has to be 2D) ex. [[1, 1], [2, 2]]
        :param y_train: labels (has to be 1D) ex. [0, 1] -> (has to start from zero)
        :param accuracy_desired: Accuracy threshold dictating that training stops at that percentage.
        Enables user to allow some bias in model to prevent overfitting.
        :param learning_rate: How fast neural network learns, weights = weights - learning_rate * gradients
        :param max_iters: Max iterations (or epochs) of the model to train
        :param show_acc : whether user would like to see the accuracy in training
        :return:
        """

        x_train, y_train = _preprocess_data(x_train, y_train)

        def param_init(x_train):
            '''initialize weights and biases'''
            num_features = len(x_train[0])  # number of features
            weights = np.random.randn(num_features, 1)
            bias = np.random.uniform()
            return weights, bias

        weights, bias = param_init(x_train)
        self.num_features = len(x_train[0])

        m = len(y_train)  # useful for derivatives in training

        iterations = tqdm(range(max_iters))
        for iteration in iterations:
            """"
            Training model. Stops when the accuracy_desired is met or too many iterations. 
            Specifications about the model : 
            ->> uses MSE loss
            ->> uses gradient descent

            In forward-propagation : 
            ->> y_hat (predictions) = weights * inputs + bias 
            ->> one_liner : y_hat = np.dot(x_train, weights) + bias 

            In gradient descent : 
            ->> Uses MSE (mean-squared-error) derivative of (yH - y) / m 
            ->> weights gradients in the inputs transpose (dot) derivative of MSE loss
            ->> bias gradients are just the sum of the errors

            Evaluation : 
            ->> Measures accuracy by seeing the error_rmse (how much it is off from the label on average)
            and then doing 1 - (error/rmse / np.mean(np.abs(y_train)) to see whether error_rmse is a lot or little given 
            the scale of the data.

            """

            y_hat = np.dot(x_train, weights) + bias  # forward pass, y_hat = wx + b

            # gradient descent
            dLdYh = (y_hat - y_train) / m  # MSE derivative
            weights -= learning_rate * np.dot(x_train.T, dLdYh)  # multiply by x_train.T as ∂Yh/∂W is just x_train
            bias -= learning_rate * sum(dLdYh)

            error_rmse = np.sqrt(np.sum((y_hat - y_train) ** 2) / (
                    2 * m))  # RMSE - how far on average are my predictions from their targets?

            accuracy_perc = 1 - (error_rmse / np.mean(np.abs(y_train)))

            if show_acc:
                iterations.set_description("accuracy : " + str(round(accuracy_perc * 100, 2)) + "% ")

            if accuracy_perc >= accuracy_desired:
                iterations.close()
                print("Accuracy threshold met!")
                break

            self.weights = weights  # updated every single iteration so you can manually stop training and not lose training progress
            self.bias = bias

    def predict(self, prediction_data):
        """
        :param prediction_data: points to be predicted on. Has to be stored in 2D array, even if just one data point. Ex. : [[1, 1]] or [[1, 1], [2, 2], [3, 3]]
        :return: flattened numpy array of your data going through the forward pass
        """

        prediction_data = np.array(prediction_data)
        if len(prediction_data.shape) < 2:
            prediction_data = np.array(
                [np.array([x_i]) for x_i in prediction_data])  # simple preprocessing to make it 2-Dimensional

        return (np.dot(prediction_data, self.weights) + self.bias).flatten()  # go through the forward pass for this
        # prediction_data

    def evaluate(self, x_test, y_test):
        """
        :param x_test: data to be evaluated on
        :param y_test: labels
        :return: score

        Evaluation done same as in training, calculate how far off from target on average (error_rmse), and how much that is compared to scale of data (np.mean(np.abs(y_test))
        """

        x_test, y_test = _preprocess_data(x_test, y_test)

        y_pred = []  # y_pred stores predictions

        for data in x_test:
            # forward pass for all data
            y_pred.append(np.dot(data, self.weights) + self.bias)

        y_pred = np.array(y_pred).flatten()
        y_test = np.array(
            y_test).flatten()  # make sure y_pred & y_test are same shape so y_pred - y_test works. Use the .flatten() method for this.
        m = len(y_pred)  # use for averaging evaluation
        error_rmse = np.sqrt(
            np.sum((y_pred - y_test) ** 2) / (2 * m))  # how much am are predictions off from targets on average?
        return 1 - (error_rmse / np.mean(np.abs(y_test)))  # how does that compare to the scale of the data

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions given by model
        :param y_test: actual labels

        visualize data using matplotlib.pyplot ->> (only for 2D data)
        """
        import matplotlib.pyplot as plt
        plt.cla()
        plt.plot([_ for _ in range(len(y_pred))], y_pred, color="blue",
                 label="predictions/y_pred")  # plot all predictions in blue
        plt.plot([_ for _ in range(len(y_test))], y_test, color="green",
                 label="labels/y_test")  # plot all labels in green
        plt.legend()
        plt.show()


def sigmoid(x):
    '''sigmoid/logistic function in Logistic Regression.'''
    return 1 / (1 + np.exp(-x))


class LogisticRegression():
    def __init__(self):
        self.information = '''

                Logistic Regression, contrary to its name, is actually a classification algorithm. This 
            means instead of trying to predict numbers, it will predict whether something is a dog or a cat for instance. 
            This algorithm still tries to make a line, just of a different shape. The use of sigmoid (1/ (1 + e^-x)) 
            function over all inputs helps create that shape optimal for binary classification. Logistic Regression can 
            classify data into 2 (binary) categories. No changes from the standard linear regression module in calls. For 
            classification of ≥ 2 classes, softmax regression is the way to go! 

            © Anish Lakkapragada 2020
        '''

    def reset_params(self):
        '''when model predictions aren't good you can always do model.reset_params() and retrain. Always train before resetting parameters.'''
        self.weights = np.random.randn(self.num_features, 1)
        self.bias = np.random.uniform()

    def fit(self, x_train, y_train, accuracy_desired=0.95, learning_rate=0.01, max_iters=10000, show_acc=True):
        """
        :param x_train: input (has to be 2D) ex. [[1, 1], [2, 2]]
        :param y_train: labels (has to be 1D) ex. [0, 1] -> (has to start from zero)
        :param accuracy_desired: Accuracy threshold dictating that training stops at that percentage.
        Enables user to allow some bias in model to prevent overfitting.
        :param learning_rate: How fast neural network learns, weights = weights - learning_rate * gradients
        :param max_iters: Max iterations (or epochs) of the model to train
        :param show_acc : whether user would like to see the accuracy in training
        :return:
        """

        x_train, y_train = _preprocess_data(x_train, y_train)

        def param_init(x_train):
            '''initialize weights and biases'''
            num_features = len(x_train[0])  # number of features
            weights = np.random.randn(num_features, 1)
            bias = np.random.uniform()
            return weights, bias

        weights, bias = param_init(x_train)
        self.num_features = len(x_train[0])
        m = len(y_train)
        iterations = tqdm(range(max_iters))
        for iteration in iterations:
            """"
            Training model. Stops when the accuracy_desired is met or too many iterations. 
            Specifications about the model : 
            ->> uses log loss (a.k.a binary cross entropy)  = -1/m * ∑ y(i)log(yH(i)) + (1-y(i))log(1-yH(i))
            ->> derivative of loss with respect to Z, where yH = sigmoid(Z) and Z = wx + b, will be 1/m * yH - y due to use of log loss
            ->> uses gradient descent

            In forward-propagation : 
            ->> z = weights * inputs + bias (note : z never actually defined just 1 line)
            ->> y_hat (predictions) = sigmoid(z)
            ->> one_liner : y_hat = sigmoid(np.dot(x_train, weights) + bias)

            In gradient descent : 
            ->> Same derivative as linear regression due to using log loss (instead of MSE)
            ->> weights gradients in the inputs transpose (dot) derivative of log loss wrt. z
            ->> bias gradients are just the sum of the dLdZ

            Evaluation : 
            ->> Measures accuracy very simple, just what percent of data is predicted in the correct class? Stops
            when accuracy_desired is met.

            """
            y_hat = sigmoid(np.dot(x_train, weights) + bias)  # forward pass, y_hat = sigmoid(wx + b)

            # gradient descent
            y_hat, y_train = y_hat.reshape(m, 1), y_train.reshape(m, 1)
            dLdZ = (y_hat - y_train) / m  # Log Loss derivative
            weights -= learning_rate * np.dot(x_train.T, dLdZ)
            bias -= learning_rate * sum(dLdZ)

            num_correct = sum([1 if pred == label else 0 for pred, label in
                               zip(np.round_(y_hat), y_train)])  # what percent are labeled correctly - super simple!

            accuracy_perc = num_correct / m

            if show_acc:
                iterations.set_description("accuracy : " + str(round(accuracy_perc * 100, 2)) + "% ")

            if accuracy_perc >= accuracy_desired:
                iterations.close()
                print("Accuracy threshold met!")
                break

            self.weights = weights
            self.bias = bias

    def predict(self, prediction_data):
        """
        :param prediction_data: points to be predicted on. Has to be stored in 2D array, even if just one data point. Ex. : [[1, 1]] or [[1, 1], [2, 2], [3, 3]]
        :return: flattened numpy array of your data going through the forward pass (that's rounded as it's either 0 or 1)
        """
        prediction_data = np.array(prediction_data)
        if len(prediction_data.shape) < 2:
            prediction_data = np.array([np.array([x_i]) for x_i in prediction_data])  # make it 2D

        return np.round_(sigmoid(np.dot(prediction_data, self.weights) + self.bias)).flatten()

    def evaluate(self, x_test, y_test):
        """
        :param x_test: data to be evaluated on
        :param y_test: labels
        :return: score

        Evaluation, just like in training, is the percent of data that is predicted to the correct class.
        """
        x_test, y_test = _preprocess_data(x_test, y_test)

        y_pred = []  # stores predictions

        for data in zip(x_test):
            y_pred.append(sigmoid(np.dot(data, self.weights) + self.bias))

        y_pred = np.array(y_pred).flatten()
        y_test = np.array(y_test).flatten()  # make sure same shape
        num_correct = sum([1 if pred == label else 0 for pred, label in
                           zip(np.round_(y_pred), y_test)])  # what percent properly predicted?
        return num_correct * 100 / len(y_test)

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions given by model
        :param y_test: actual labels

        visualize data using matplotlib.pyplot ->> (only for 2D data)
        """
        import matplotlib.pyplot as plt
        plt.cla()
        y_pred, y_test = y_pred.flatten(), y_test.flatten()
        plt.scatter([_ for _ in range(len(y_pred))], y_pred, color="blue", label="predictions/y_pred")
        plt.scatter([_ for _ in range(len(y_test))], y_test, color="green", label="labels/y_test")
        plt.legend()
        plt.show()


def softmax(scores):
    '''
    Used for Softmax Regression in the forward pass.

    :param scores: pre-softmax outputs for softmax regression
    :return: return each element's exponential divided by some of all element's exponential
    '''
    return np.exp(scores) / np.sum(np.exp(scores))


class SoftmaxRegression():
    def __init__(self, **kwargs):
        self.information = '''

                    Softmax Regression is a classification algorithm for when you're classes are >= 2 (not binary). This 
                makes it a versatile for datasets like Iris where there are many classes. Softmax Regression builds 
                off of Logistic Regression (binary classifier) so be familiar with that prior using this algorithm. 
                However, it is worth noting that Softmax Regression performs the same as Logistic Regression when 
                there are just 2 classes (binary.) In the fit() method of this class make sure to have the 
                "num_classes" keyword , for instance : model.fit(data, labels, num_classes = ?) This is needed for 
                the algorithm to learn. 

                © Anish Lakkapragada 2020 

        '''

    def reset_params(self):
        '''when model predictions aren't good you can always do model.reset_params() and retrain. Always train before resetting parameters.'''
        self.weights = np.random.randn(self.num_features, self.num_classes)
        self.bias = np.random.uniform()

    def fit(self, x_train, y_train, num_classes, accuracy_desired=0.95, learning_rate=0.01, max_iters=10000,
            show_acc=True):
        """
        :param x_train: input (has to be 2D) ex. [[1, 1], [2, 2]]
        :param y_train: labels (has to be 1D) ex. [0, 1] -> (has to start from zero)
        :param accuracy_desired: Accuracy threshold dictating that training stops at that percentage.
        Enables user to allow some bias in model to prevent overfitting.
        :param learning_rate: How fast neural network learns, weights = weights - learning_rate * gradients
        :param max_iters: Max iterations (or epochs) of the model to train
        :param show_acc : whether user would like to see the accuracy in training
        :return:
        """

        x_train, _ = _preprocess_data(x_train,
                                      y_train)  # the other labels don't matter because y_train has to be one-hot-encoded
        y_train = one_hot(y_train, num_classes=num_classes)  # one_hot_encode the labels
        x_train, y_train = np.array(x_train), np.array(y_train)  # convert to numpy arrays

        def param_init(x_train):
            num_features = len(x_train[0])
            weights = np.random.randn(num_features, num_classes)
            bias = np.random.randn(1, num_classes)
            return weights, bias

        weights, bias = param_init(x_train)
        self.num_features = len(x_train[0])
        self.num_classes = num_classes  # important for parameter intializations
        m = len(y_train)

        iterations = tqdm(range(max_iters))
        for iteration in iterations:
            """
            Training model. Stops when the accuracy_desired is met or too many iterations. 
            Specifications about the model : 
            ->> uses categorical cross entropy loss (-1/m * ∑y * ln(y_hat))
            ->> keep in mind that you do softmax for all scores in y_hat, so not a one-liner
            ->> bias is now the number of classes, so more verbose calculations there


            In forward-propagation : 
            ->> z = weights * inputs + bias (note : z = np.dot(x_train, weights) + bias)
            ->> y_hat (predictions) = [softmax(z_i) for z_i in z]
            ->> first two lines of forward pass

            In gradient descent : 
            ->> Same derivative as logistic regression due to using cross entropy loss (same as binary cross entropy derivative)
            ->> weights gradients in the inputs transpose (dot) derivative of crossentropy loss wrt. z
            ->> bias gradients are just the sum of the dLdZ

            Evaluation : 
            ->> Measures accuracy very simple, just what percent of data is predicted in the correct class? Stops
            when accuracy_desired is met.  
            """

            z = np.dot(x_train, weights) + bias  # forward pass start
            y_hat = np.array([softmax(z_i) for z_i in z])  # softmax for each output

            # gradient descent
            dLdZ = (y_hat - y_train) / m  # crossentropy loss derivative
            weights -= learning_rate * np.dot(x_train.T, dLdZ)
            for error in dLdZ: bias -= learning_rate * error  # change bias -> dLdZ is a matrix now due to one_hot_labels
            num_correct = sum([1 if np.argmax(pred) == np.argmax(label) else 0 for pred, label in
                               zip(np.round_(y_hat), y_train)])  # what percent are labeled correctly - super simple!

            accuracy_perc = num_correct / m
            if show_acc:
                iterations.set_description("accuracy : " + str(round(accuracy_perc * 100, 2)) + "% ")

            if accuracy_perc >= accuracy_desired:
                iterations.close()
                print("Accuracy threshold met!")
                break

            self.weights = weights
            self.bias = bias

    def predict(self, prediction_data):
        """
        :param prediction_data: points to be predicted on. Has to be stored in 2D array, even if just one data point. Ex. : [[1, 1]] or [[1, 1], [2, 2], [3, 3]]
        :return: flattened numpy array of your data going through the forward pass (that's rounded as it's either 0 or 1)
        """

        prediction_data = np.array(prediction_data)
        if len(prediction_data.shape) < 2:
            prediction_data = np.array([np.array([x_i]) for x_i in prediction_data])  # make it 2D

        Z = np.dot(prediction_data, self.weights) + self.bias
        y_hat = np.array([softmax(z) for z in Z])
        return np.array([np.argmax(output) for output in y_hat])  # go through each output and give the index of where
        # the probability is highest for that output
        # ex. [[0.2, 0.3, 0.5], [0.9, 0.06, 0.04]] will become [2, 0]

    def evaluate(self, x_test, y_test):
        """
        :param x_test: data to be evaluated on
        :param y_test: labels
        :return: score

        Evaluation, just like in training, is the percent of data that is predicted to the correct class.
        """
        x_test, _ = _preprocess_data(x_test, y_test)
        y_test = one_hot(y_test, self.num_classes)

        Z = np.dot(x_test, self.weights) + self.bias
        y_pred = np.array([softmax(z) for z in Z])
        y_pred = np.array(y_pred)
        num_correct = sum(
            [1 if np.argmax(pred) == np.argmax(label) else 0 for pred, label in zip(np.round_(y_pred), y_test)])
        return num_correct * 100 / len(y_test)  # basic evaluation

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions given by model
        :param y_test: actual labels

        visualize data using matplotlib.pyplot ->> (only for 2D data)
        """
        import matplotlib.pyplot as plt
        plt.cla()
        y_pred, y_test = y_pred.flatten(), y_test.flatten()
        plt.scatter([_ for _ in range(len(y_pred))], y_pred, color="blue", label="predictions/y_pred")
        plt.scatter([_ for _ in range(len(y_test))], y_test, color="green", label="labels/y_test")
        plt.legend()
        plt.show()


def _poly_transform(data):
    '''take data and transform it like [2, 3, 4] - > [2, 9, 64]'''
    new_data = []
    for observation in data:
        new_observation = []
        for feature in range(len(observation)): new_observation.append(math.pow(observation[feature], (feature + 1)))
        new_data.append(new_observation)
    return np.array(new_data)


class PolynomialRegression():
    '''
    Polynomial Regression here works by first :
    1. taking data in
    2. transforming it into new data, where every element in the data ^ (index_in_observation + 1).
    For example : [[1, 2, 3], [2, 3, 4]] -> [[1, 4, 27], [2, 9, 64]]
    '''

    def __init__(self):
        self.information = """

            Polynomial Regression is an extension of Linear Regression. Just like linear regression 
            it predicts numbers, not categories. The only difference is that each observation/input in your data has each 
            of its elements go to the power of its num_feature in the data. If that doesn't make any sense that's okay. 
            Here's an example. 

            [2, 3, 4] -> where 2 is the 1th feature, 3 is the 2nd feature, 4 is the 3rd feature. 
            Therefore transforming the data though function t(x), gives this : t([2, 3, 4]) = [2, 9, 64]
            That's because 2^1 = 2, 3^2 = 9 , 4^3 = 64.

            These transformed inputs are then learnt by the Linear Regression Algorithm. 
        """
        self.inner_linear_model = None  # this inner_linear_model is the Linear Regression model used to predict

    def reset_params(self):
        self.inner_linear_model.reset_params()  # calls the LinearRegression.reset_params() method

    def fit(self, x_train, y_train, accuracy_desired=0.95, learning_rate=0.01, max_iters=10000, show_acc=True):
        """
        :param x_train: input (has to be 2D) ex. [[1, 1], [2, 2]]
        :param y_train: labels (has to be 1D) ex. [0, 1] -> (has to start from zero)
        :param accuracy_desired: Accuracy threshold dictating that training stops at that percentage.
        Enables user to allow some bias in model to prevent overfitting.
        :param learning_rate: How fast neural network learns, weights = weights - learning_rate * gradients in gradient descent
        :param max_iters: Max iterations (or epochs) of the model to train
        :param show_acc : whether user would like to see the accuracy in training
        :return:
        """

        x_train, y_train = _preprocess_data(x_train, y_train)
        x_train = _poly_transform(x_train)  # transform data

        self.inner_linear_model = LinearRegression()  # build the linear regression model
        self.inner_linear_model.fit(x_train, y_train, accuracy_desired=accuracy_desired, learning_rate=learning_rate,
                                    max_iters=max_iters, show_acc=show_acc)  # fit it (using all given parameters)

    def predict(self, prediction_data):
        """
        :param prediction_data: points to be predicted on. Has to be stored in 2D array, even if just one data point. Ex. : [[1, 1]] or [[1, 1], [2, 2], [3, 3]]
        :return: flattened numpy array of your data going through the forward pass (that's rounded as it's either 0 or 1)
        """
        prediction_data = _poly_transform(prediction_data)
        return self.inner_linear_model.predict(prediction_data)  # just use the prediction method

    def evaluate(self, x_test, y_test):
        """
        :param x_test: data to be evaluated on
        :param y_test: labels
        :return: score

        Evaluation using the same method as Linear Regression.
        """
        x_test = _poly_transform(x_test)
        return self.inner_linear_model.evaluate(x_test, y_test)

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions given by model
        :param y_test: actual labels

        visualize data using matplotlib.pyplot ->> (only for 2D data)
        """
        import matplotlib.pyplot as plt
        plt.cla()
        plt.plot([_ for _ in range(len(y_pred))], y_pred, color="blue", label="predictions/y_pred")
        plt.plot([_ for _ in range(len(y_test))], y_test, color="green", label="labels/y_test")
        plt.legend()
        plt.show()


def reLU(X, leaking=0):
    """
    :param x: Input
    :param leaking: Leaking value, default 0 for vanilla reLU activation
    :return:
    """
    return np.array([x_i if x_i >= 0 else x_i * leaking for x_i in X])


def reLU_deriv(y_hat, leaking=0):
    """
    :param y_hat:
    :return: reLU derivative of each element in y_hat
    """
    return np.array([[1] if pred > 0 else [leaking] for pred in y_hat])


class ReLURegression:
    def __init__(self):
        self.information = "dwadwa"

    def reset_params(self):
        '''when model predictions aren't good you can always do model.reset_params() and retrain. Always train before resetting parameters.'''
        self.weights = np.random.randn(self.num_features, 1)
        self.bias = np.random.uniform()

    def fit(self, x_train, y_train, leaking=0, accuracy_desired=0.95, learning_rate=0.01, max_iters=10000,
            show_acc=True):
        """
        :param x_train: input (has to be 2D) ex. [[1, 1], [2, 2]]
        :param y_train: labels (has to be 1D) ex. [0, 1] -> (has to start from zero)
        :param accuracy_desired: Accuracy threshold dictating that training stops at that percentage.
        Enables user to allow some bias in model to prevent overfitting.
        :param learning_rate: How fast neural network learns, weights = weights - learning_rate * gradients
        :param max_iters: Max iterations (or epochs) of the model to train
        :param show_acc : whether user would like to see the accuracy in training
        :return:
        """

        x_train, y_train = _preprocess_data(x_train, y_train)  # first preprocess data
        self.leaking = leaking

        def param_init(x_train):
            num_features = len(x_train[0])
            weights = np.random.randn(num_features, 1)
            bias = np.random.uniform()
            return weights, bias

        weights, bias = param_init(x_train)
        self.num_features = len(x_train[0])
        m = len(y_train)
        iterations = tqdm(range(max_iters))
        for iteration in iterations:
            """"
            Training model. Stops when the accuracy_desired is met or too many iterations. 
            Specifications about the model : 
            ->> uses MSE loss, simple derivative of (yH - y)/m
            ->> uses (leaking) relu activation function, which is x if x >= 0 or leaking_param * x if x < 0 
            ->> uses gradient descent

            In forward-propagation : 
            ->> Z = wx + b
            ->> y_hat = (leaking) reLU(Z)  (NOTE : Z never defined)
            ->> one_liner = reLU(np.dot(x_train, weights) + bias)

            In gradient descent : 
            ->> Same derivative as linear regression for ∂L/∂Yh which is (yH - y)/,
            ->> However, that vector needs to go through the reLU derivative (which is if yH > 0 : 1,  else : leaking)
            ->> multiply the two together to get dLdZ

            Evaluation : 
            ->> Measures accuracy same as linear regression. Check docs above.
            """

            y_hat = reLU(np.dot(x_train, weights) + bias, leaking=leaking)  # forward pass, y_hat = relu(Z) | Z = wx + b

            # gradient descent
            y_hat, y_train = y_hat.reshape(m, 1), y_train.reshape(m, 1)
            dLdYh = (y_hat - y_train) / m
            dYhdZ = reLU_deriv(y_hat, leaking=leaking)
            dLdZ = dLdYh * dYhdZ  # chain rule!
            weights -= learning_rate * np.dot(x_train.T, dLdZ)
            bias -= learning_rate * np.sum(dLdZ)

            error_rmse = np.sqrt(np.sum((y_hat - y_train) ** 2) / (
                    2 * m))  # RMSE - how far on average are my predictions from their targets?

            accuracy_perc = 1 - (error_rmse / np.mean(np.abs(y_train)))

            if show_acc:
                iterations.set_description("accuracy : " + str(round(accuracy_perc * 100, 2)) + "% ")

            if accuracy_perc >= accuracy_desired:
                iterations.close()
                print("Accuracy threshold met!")
                break

            self.weights = weights  # updated every single iteration so you can manually stop training and not lose training progress
            self.bias = bias

    def predict(self, prediction_data):
        """
        :param prediction_data: points to be predicted on. Has to be stored in 2D array, even if just one data point. Ex. : [[1, 1]] or [[1, 1], [2, 2], [3, 3]]
        :return: flattened numpy array of your data going through the forward pass
        """

        prediction_data = np.array(prediction_data)
        if len(prediction_data.shape) < 2:
            prediction_data = np.array(
                [np.array([x_i]) for x_i in prediction_data])  # simple preprocessing to make it 2-Dimensional

        return reLU(np.dot(prediction_data, self.weights) + self.bias,
                    leaking=self.leaking).flatten()  # go through the forward pass for this
        # prediction_data

    def evaluate(self, x_test, y_test):
        """
        :param x_test: data to be evaluated on
        :param y_test: labels
        :return: score

        Evaluation done same as in training, calculate how far off from target on average (error_rmse), and how much that is compared to scale of data (np.mean(np.abs(y_test))
        """

        x_test, y_test = _preprocess_data(x_test, y_test)

        y_pred = []  # y_pred stores predictions

        for data in x_test:
            # forward pass for all data
            y_pred.append(reLU(np.dot(data, self.weights) + self.bias, leaking=self.leaking))

        y_pred = np.array(y_pred).flatten()
        y_test = np.array(
            y_test).flatten()  # make sure y_pred & y_test are same shape so y_pred - y_test works. Use the .flatten() method for this.
        m = len(y_pred)  # use for averaging evaluation
        error_rmse = np.sqrt(
            np.sum((y_pred - y_test) ** 2) / (2 * m))  # how much am are predictions off from targets on average?
        return 1 - (error_rmse / np.mean(np.abs(y_test)))  # how does that compare to the scale of the data

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions given by model
        :param y_test: actual labels

        visualize data using matplotlib.pyplot ->> (only for 2D data)
        """
        import matplotlib.pyplot as plt
        plt.cla()
        plt.plot([_ for _ in range(len(y_pred))], y_pred, color="blue",
                 label="predictions/y_pred")  # plot all predictions in blue
        plt.plot([_ for _ in range(len(y_test))], y_test, color="green",
                 label="labels/y_test")  # plot all labels in green
        plt.legend()
        plt.show()

log_reg = LogisticRegression()
log_reg.fit(x_train, y_train)
print(time.time() - start_time)


"""
This is the neural networks subpackage of SeaLion for building, training, and using deep neural networks. Enjoy!
"""

from . import layers
from . import loss
from . import models
from . import optimizers
from . import utils_nn


import math, random
import numpy as np
def get_new_data(x_train, y_train, k_value=5, num_new_data=20) :

    def euclidean_distance(point1, point2):
        sum = 0
        for x in range(len(point1)):
            sum += (point2[x] - point1[x]) ** 2
        sum = math.sqrt(sum)
        return sum

    all_points = {}  # point : label
    for point, label in zip(x_train, y_train):
        all_points[tuple(point)] = label
    # we now need to get the maximum of everycoordinate
    num_dimensions = len(x_train[0])
    maximum_coordinates = []
    for _ in range(num_dimensions):
        maximum_coordinates.append(0)
    minimum_coordinates = []
    for _ in range(num_dimensions)  :
        minimum_coordinates.append(0)
    for point in x_train:
        for cor in range(len(point)):
            if maximum_coordinates[cor] < point[cor]:
                maximum_coordinates[cor] = point[cor]
            if minimum_coordinates[cor] > point[cor] :
                minimum_coordinates[cor] = point[cor]
   # given_desired_new_data_num = 0
    desired_new_data_num = num_new_data
    k_value = 9
    new_points = {}

    for _ in range(desired_new_data_num):
        # the first step is create a new point
        new_point = []
        for x in range(num_dimensions):
            new_point.append(random.uniform(minimum_coordinates[x], maximum_coordinates[x]))

        # print("This is the new_point : ", new_point)

        # the knn_model thing we have to do is go through all the points
        distances_to_points = {}  # distance : point
        for point in all_points:
            distance = euclidean_distance(point, new_point)
            # take this distance and attach it to another dictionary to store it
            distances_to_points[distance] = point

        # right now we need to find the closest points, or the ones with the lowest distances
        sorted_distances_to_points_keys = sorted(distances_to_points)
        sorted_distances_to_points_keys = sorted_distances_to_points_keys[:k_value]
        # now we have the closest distances time to get the closest points
        closest_points = []
        for distance in sorted_distances_to_points_keys:
            closest_points.append(distances_to_points[distance])

        # now that we have the closest points - we need to get their labels from the al_points dict
        closest_points_labels = []
        for point in closest_points:
            closest_points_labels.append(all_points[point])
        new_label = sum(closest_points_labels) / len(closest_points_labels)
        new_points[tuple(new_point)] = new_label

    X, y = [], []
    for given_point, given_label in zip(x_train, y_train):
        X.append(given_point)
        y.append(given_label)

    for point in new_points:
        X.append(list(point))
        y.append(new_points[tuple(point)])

    X, y = np.array(X), np.array(y)
    return X, y



"""
A neural network is just a composition of layers, or functions.
"""

from tensor import Tensor
import numpy as np
from layers import Layer
from typing import Sequence, Iterator, Tuple

class NeuralNet :
    def __init__(self, layers : Sequence[Layer]) -> None:
        self.layers = layers
    def forward(self, inputs : Tensor) -> Tensor:
        for layer in self.layers :
            inputs = layer.forward(inputs)
        return inputs

    def backward(self, grad : Tensor) -> Tensor:
        for layer in reversed(self.layers) :
            grad = layer.backward(grad)
        return grad

    def params_and_grads(self) -> Iterator[Tuple[Tensor, Tensor]] :
        for layer in self.layers :
            for name, param in layer.params.items() :
                grad = layer.grads[name]
                yield param, grad





'''
- data sciencetists can assume data statistics as random variables from a distribution
- a null hypothesis, H0, represents some default position and H1, the alternative hypothesis,
 is what we would like to compare it to
'''

#Example : Flipping a Coin

#null hypothesis that the coin is fair if p = 0.5, here p is the probability of landing heads

#each coin flip is a bernoulli trial, and we have binomial (n , p) random var, which can be approxcimated

from typing import Tuple
import math

def normal_approximation_to_binomial(n : int, p : float) -> Tuple[float, float] :
    '''return mu and sigma corresponding to a Binomial(n, p)'''
    mu = n * p
    sigma = math.sqrt(p * (1  - p) * n )
    return mu, sigma

#Whenever a random variable follows a normal distribution, use the cdf to see if it lies between a given interval

from DataScienceFromScratch.Probability import normal_cdf

#the Normal cdf is the probability the variable is below a threshold (cumulative)

normal_probability_below = normal_cdf

#if it's above the threshold then it's not below
def normal_probability_above(lo : float,
                             mu : float = 0,
                             sigma : float = 1) -> float :
    """the probability that an N(mu, sigma) is greater than lo"""
    return 1 - normal_probability_below(lo, mu, sigma)

#less than hi, but more than lo
def normal_probability_between(lo : float,
                               ho : float,
                               mu : float = 0,
                               sigma : float = 1) -> float :
    '''probability it's between lo and ho'''
    return normal_cdf(ho, mu, sigma) - normal_cdf(lo, mu, sigma)

#and outside (1 - between)
def normal_probability_outside(lo : float,
                               ho : float,
                               mu : float = 0,
                               sigma : float = 1) -> float :
    return 1 - normal_probability_between(lo, ho, mu, sigma)

"""
We can also do the reverse, where we find an interval which contains a given % probability

An example could be an interval with 60% probability, so a upper tail at 80% and a lower tail and 20% will leave 60%
and each will have 20% of the probability on each side. 
"""

from DataScienceFromScratch.Probability import inverse_normal_cdf

def normal_upper_bound(probability : float,
                       mu : float = 0,
                       sigma : float = 1) -> float :
    """returns the higher z, or the value that leads to the given probability on the CDF (mu, sigma)"""
    return inverse_normal_cdf(probability, mu, sigma)

def normal_lower_bound(probability : float,
                       mu : float = 0,
                       sigma : float = 1) :
    '''take a probability, like 80% and find a value giving prob 20% on the CDF'''
    return inverse_normal_cdf(1 - probability, mu, sigma)

def normal_two_sided_bounds(probability : float,
                            mu : float = 0,
                            sigma : float = 1) -> Tuple[float, float] :
    """
    Return the lower and upper bound to create an interval which contains a given probability
    """

    tail_probability = (1 - probability) / 2 #here this is 0.2

    #upper bound shuold have tail_probability above it
    upper_bound = normal_lower_bound(tail_probability, mu, sigma) #finds the upper 80% threshold

    #lower bound should have tail probability below it
    lower_bound = normal_upper_bound(tail_probability, mu, sigma) #finds the lower 20% threshold

    return lower_bound, upper_bound

mu_0, sigma_0 = normal_approximation_to_binomial(1000, 0.5) #(n,p) for our null hypothesis p = 0.5 for 1000 trials
#mu0, sigma0 if the coin is far (H0, null hypothesis is correct)

lower_bound, upper_bound = normal_two_sided_bounds(0.95, mu_0, sigma_0)
#if X, the number of heads after n trials, is outside the lower and upper bounds, then the coin is not fair (H0 is wrong)

print("Lower : ", lower_bound, " ; Upper : ", upper_bound)

#we are also want to know the Power of a test, in which we fail to reject H0 despite it being false
"""
H0 here is false if p ≠ 0.5, so we can test for p = 0.55 for example. We want to see what percent of our bound is correct
given a different p, or what's the chance we reject H0 with a different p. 
"""

lo, hi = normal_two_sided_bounds(0.95, mu_0, sigma_0) # given the actual p's mean and std, find the bounds containing
 # 95 % of the probability

#the actual mu and sigma given p = 0.55
mu_1, sigma_1 = normal_approximation_to_binomial(1000, 0.55)

# a type 2 error means we fail to reject the null hypothesis,
# which happens when X is still in the original interval for p = 0.5 (our null hypothesis.)

type2_probability = normal_probability_between(lo, hi, mu_1, sigma_1) #given the range for the null hypothesis
# find the probability shared between the two
power = 1 - type2_probability

print("This is the power : ", power)

"""
Instead if our null hypothesis was that the coin liked tails more, or p ≤ .5, and that should have X lower than 500.

hi = normal_upper_bound(0.95, mu_0, sigma_0) #upper bound for 95% confidence
type2_probability = normal_probability_below(hi, mu1, sigma1) #calculate probablity from X = 0 to the high X 
power = 1 - type2_probability #a change in the p value (giving mu1, sigma1) changes the probability on both CDFs.
"""

'''
P-values!

Compute the probability that assuming H0 is true, would we see a value at at least as extreme as observed? 

The reason we do this is because a low p-rate shows that the data assuming the null was not by chance. Therefore
we are almost calculating the significance in the alternative hypothesis with a p-value as the lower it is the less
significance. If there is less significance in the alternative hypothesis through the p-vaue then there is significance
in the null hypothesis, we reject the null hypothesis.
'''

def two_sided_p_value(x : float, mu : float = 0, sigma : float = 1) -> float :
    """How likely are we to see a value as extreme as x if values are from N(mu, sigma)?"""

    if x >= mu :
        #x is greater than the mean, the tail is everything greater than x
        return 2 * normal_probability_above(x, mu, sigma)
    else :
        return 2 * normal_probability_below(x, mu, sigma)


two_sided_p_value(529.5, mu_0, sigma_0) #0.062 or 6.5% chance that given p =0 you see 530 heads
'''we used a continuity correction for the 0 and 1 distribution (discrete) not 0 -1 distribution. So 529.5 not 530'''

#another way to prove this :

import random
extreme_value_count = 0
for _ in range(1000) :
    num_heads = sum(1 if random.random() < 0.5 else 0 #count num heads
                    for _ in range(1000)) #in 1000 flips

    if num_heads >= 530 or num_heads <= 470 : #how often is this extreme
        extreme_value_count += 1

#assert 59 <= extreme_value_count <= 65, f"{extreme_value_count}"

"""
If the p-value (sort of the significance of the alternative hypothesis) is lower than the significance of our 
null hypothesis, reject the null hypothesis.
"""

upper_p_value = normal_probability_above
lower_p_value = normal_probability_below
print("Significance : 5%")
print("525 heads : ", upper_p_value(524.5, mu_0, sigma_0), " : acccept; > significance")
print("527 heads : ", upper_p_value(526.5, mu_0, sigma_0), " : reject; < significance")

"""ALWAYS MAKE SURE THE DATA IS NORMALLY DISTRIBUTED BEFORE CALCULATING P-VALUES"""

"""
Confidence Intervals : 

They give you an interval that spans confidence_level % of the data. It is confidence_level (usually 95%) % confident 
that the parameter is there. Always create a normal distribution centered around getting that value first. 
"""

p_hat = 525 / 1000 #our estimate of how many heads
mu = p_hat #here we get mu and sigma from the 0.525 heads rate by treating each trial out of 1000 as a bernoulli trial
sigma = math.sqrt(p_hat * (1 - p_hat) / 1000)
confidence_interval = normal_two_sided_bounds(0.95, mu, sigma) #0.49 - 0.55
print("This is the confidence interval : ", confidence_interval)

#if we had a fair coin
p_hat = 0.5
mu = p_hat
sigma = math.sqrt(p_hat * (1 - p_hat) / 1000)
confidence_interval = normal_two_sided_bounds(0.95, mu, sigma) #0.46 - 0.53
print("This is the confidence interval : ", confidence_interval)

"""
p-hacking : 

You can always change the data and test enough hypothesis to get a p < 0.05 so you can reject the null hypothesis.

running it over and over again ensures that you get a hypothesis with a high significance that is likely to reject
the null hypothesis. these hypotheses generated are the experiments and the number of tails and heads in them. You are
sure to find some that are significant and that disprove the null hypothesis. That way you can find hypothesis that
do disprove the null. 
"""

from typing import List

def run_experiment() -> List[bool] :
    '''Flips a fair coin 1000 times, True is head and False is tails'''
    return [random.random() < 0.5 for _ in range(1000)]

def reject_fairness(experiment : List[bool]) -> bool :
    '''using the 5% significance level'''
    num_heads = len([flip for flip in experiment if flip]) #only if the flip is true count it
    return num_heads < 469 or num_heads > 531 #62 values are extreme in the dataset so we want to see if this is in
    # the fair boundary

random.seed(0)
experiments = [run_experiment() for _ in range(1000)]
num_rejections = len([experiment
                      for experiment in experiments
                      if reject_fairness(experiment)]) #count the number of rejected experiments
assert num_rejections == 46

print("we are rejecting ", num_rejections, " points of the data, not 62 like with the null hypothesis - "
                                            "therefore bye null!")


"""
Running A/B test. 

Let's say Na people see ad A, and na click it. Each ad is a Bernoulli trial where pa is the probably 
somebody clicks it. 

That means we can represent the trials mean as pa and its standard deviation as sigmaA = sqrt(pa(1-pa) / Na)

Similarity ad B has mean pb and std sigmaB = sqrt(pb(1 - pb) / Nb)
"""

def estimated_parameters(N : int, n : int) -> Tuple[float, float] :
    p = n / N
    sigma = math.sqrt(p * (1 - p) / N)
    return p, sigma #what percent click it, std

#our null hypothesis is that pa = pb

def a_b_test_statistic(N_a : int, n_A : int, N_b : int, n_B : int) -> float :
    p_A, sigma_A = estimated_parameters(N_a, n_A)
    p_B, sigma_B = estimated_parameters(N_b, n_B)
    return (p_B - p_A) / math.sqrt(sigma_A ** 2 + sigma_B ** 2) #todo understand why this?

z = a_b_test_statistic(1000, 200, 1000, 180) #ad A got 200 / 1000 clicks and ad B got 180 / 1000 clicks
print("probability of change/importance : ", two_sided_p_value(z)) #todo understand why p-value?

z = a_b_test_statistic(1000, 200, 1000, 150)
print("probability of change/importance : ", two_sided_p_value(z)) #barely anything

"""
Bayesian Inference : 

Here you first start out with a prior distribution and use bayes theorem to update based on new data
the chance of any event happening (this updated distribution is the posterior distribution.)  

You use a beta distribution to put the probabilities from 0 to 1 to create such a distribution. 

An example would first to start out with Beta(20, 20) which expresses the belief of a fair coin. Then flip a bit more, 
count the number of heads and tails and update beta as beta + t and alpha as alpha + h where h, t are the number
of heads and tails respectively. You may get Beta(100, 101) where alpha = 100, beta = 101 - which is very true to 
our original hypothesis. Bayesian Inference is about updating past beliefs on the evidence found in new data. 
"""

def B(alpha : float, beta : float) -> float :
    '''Normalizing factor makes sure probabilities are from 0 - 1'''
    return math.gamma(alpha) * math.gamma(beta) / math.gamma(alpha + beta)

def beta_pdf(x : float, alpha : float, beta : float) -> float :
    if x <= 0 or x >= 1 : #outside of pdf
        return 0
    return x ** (alpha - 1) * (1 - x) ** (beta - 1) / B(alpha, beta)

#the distribution is centered at alpha / (alpha + beta)


from sealion.ensemble_learning import EnsembleClassifier
from sealion.decision_trees import DecisionTree
from regression_gd import SoftmaxRegression
from sealion.regression import LogisticRegression

log_reg = LogisticRegression()
softmax = SoftmaxRegression(2)
dt = DecisionTree()

import numpy as np
import pandas as pd
data = pd.read_csv('/users/anish/downloads/breast-cancer-wisconsin.data')

def data_collection(data) :
    training_data = data[['ID', 'clump_thickness','cell_uniformity', 'shape_uniformity', 'chromatin','nucleoli']]
    labels = data['class']

    training_data = training_data.drop(['ID', 'chromatin'], axis = 1)
    from sklearn.preprocessing import StandardScaler
    scaler = StandardScaler()
    training_data = training_data._get_numeric_data()

    training_data = scaler.fit_transform(training_data)

    training_data, labels = np.array(training_data), np.array(labels)

    split = 30
    x_test, y_test = training_data[:split], labels[:split]
    x_train,y_train = training_data[split:], labels[split:]

    x_train, y_train, x_test, y_test = np.array(x_train, dtype = float), np.array(y_train, dtype = float), np.array(x_test, dtype = float), np.array(y_test, dtype = float)
    y_train, y_test = y_train/2 - 1, y_test/2 - 1

    return x_train,x_test, y_train, y_test

x_train, x_test, y_train, y_test = data_collection(data)

ec = EnsembleClassifier({'sfmx' : softmax, "log" : log_reg, "dt" : log_reg}, classification = True)
ec.fit(x_train, y_train)
ec.evaluate(x_test, y_test)



import numpy as np
cimport numpy as np
from collections import Counter

cdef np.ndarray[np.float64_t, ndim = 1] _probabilties(float feature, dict label_feature_counts, dict num_class_data) :
    cdef list probs = []
    cdef int label
    cdef float prob
    for label in label_feature_counts :
        prob = (label_feature_counts[label][feature] + 1e-5) / (num_class_data[label] + 2e-5) # (num_times_in_that_label + e) / (num_label + 2e)
        probs.append(prob)
    return np.array(probs)

cdef class cy_MultinomialNaiveBayes():
    cdef dict num_class_data, label_feature_counts
    cdef set labels_set

    cpdef fit(self, x_train, y_train):
        x_train, y_train = np.array(x_train), np.array(y_train)
        if x_train.dtype == np.dtype('O') : raise ValueError("Expects same number of features in each row of data.")

        cdef list labels_set = y_train.tolist() #possible labels
        labels_set.sort()

        self.labels_set = set(labels_set)
        self.num_class_data = {lbl : np.count_nonzero(y_train == lbl) for lbl in list(self.labels_set)}

        self.label_feature_counts = {}

        cdef int label
        cdef np.ndarray data_for_label

        for label in list(self.labels_set) :
            data_for_label = x_train[np.where(y_train == label)]
            data_for_label = data_for_label.flatten().flatten().flatten()
            self.label_feature_counts[label] = Counter(data_for_label)  #for each label, give me how common each feature is


    cpdef list predict(self, x_test) :
        cdef list predictions = []
        x_test = np.array(x_test, dtype = np.float64)

        #some declarations for the for-loop
        cdef np.ndarray[np.float64_t, ndim = 1] current_probs, pred_data
        cdef float feature

        for pred_data in x_test :
            current_probs = np.zeros(len(self.labels_set), dtype = np.float64)
            for feature in pred_data :
                current_probs += np.log(_probabilties(feature, self.label_feature_counts, self.num_class_data))
            current_probs = np.exp(current_probs)
            predictions.append(list(self.labels_set)[np.argmax(current_probs)])
        return predictions

    cpdef float evaluate(self, x_test, y_test) :
        cdef list y_pred = cy_MultinomialNaiveBayes.predict(self, x_test)
        return sum([1 if pred == label else 0 for pred, label in zip(y_pred, y_test)]) / len(y_test)

def _normal_distribution(value, sigma_and_mu) :
    sigma, mu = sigma_and_mu
    return 1 / (sigma * np.sqrt ( 2 * np.pi)) * np.exp(-0.5 * np.power( (value - mu) / sigma , 2))

cdef class cy_GaussianNaiveBayes() :
    cdef set labels_set
    cdef int num_data
    cdef dict num_class_data, labels_to_feature_distribs

    cpdef void fit(self, x_train, y_train) :
        '''everything is non-numeric'''

        x_train = np.array(x_train)
        y_train = np.array(y_train)

        if x_train.dtype == np.dtype('O') : raise ValueError("Expects same number of features in each row of data.")

        cdef list labels_set = y_train.tolist()  # possible labels
        labels_set.sort()
        self.labels_set = set(labels_set)

        cdef float lbl
        self.num_class_data = {lbl: np.count_nonzero(y_train == lbl) for lbl in self.labels_set}
        self.num_data = len(x_train)

        self.labels_to_feature_distribs = {}
        cdef int feature
        cdef dict feature_dict
        cdef np.ndarray dataset
        cdef float label

        for label in list(self.labels_set) :
            feature_dict = {}
            for feature in range(len(x_train[0])) :
                dataset = x_train[np.where(y_train == label)][:, feature]
                feature_dict[feature] = [np.std(dataset), np.mean(dataset)]
            self.labels_to_feature_distribs[label] = feature_dict

    cpdef list predict(self, x_test):
        cdef list predictions = []
        x_test = np.array(x_test, dtype = np.float64)

        cdef np.ndarray[np.float64_t, ndim = 1] pred_data
        cdef list probs, sigma_and_mu
        cdef int label, feature
        cdef dict feature_dict
        cdef float prob_label

        for pred_data in x_test :
            probs = []
            for label in list(self.labels_set) :
                feature_dict = self.labels_to_feature_distribs[label]
                prob_label = np.log(self.num_class_data[label] / self.num_data)
                for feature, sigma_and_mu in feature_dict.items() :
                    prob_label += np.log(_normal_distribution(pred_data[feature], sigma_and_mu) + 1e-29)
                probs.append(np.exp(prob_label)) # needs the exp
            predictions.append(list(self.labels_set)[np.argmax(np.array(probs))])

        return predictions

    cpdef float evaluate(self, x_test, y_test) :
        cdef list y_pred = cy_GaussianNaiveBayes.predict(self, x_test)
        return sum([1 if pred == label else 0 for pred, label in zip(y_pred, y_test)]) / len(y_test)



"""
We use an optimizer to adjust the parameters of our network based on the gradients
computed in backpropagation
"""

from nn import NeuralNet

class Optimizer :
    def step(self, net : NeuralNet) -> None:
        raise NotImplementedError

class SGD(Optimizer) :
    def __init__(self, lr : float = 0.01) -> None:
        self.lr = lr

    def step(self, net : NeuralNet) -> None:
        for param, grad in net.params_and_grads() :
            param -= self.lr * grad


import setuptools
from distutils.core import setup

non_python_files = ['cython_ensemble_learning.cpython-38-darwin.so', 'cython_decision_tree_functions.cpython-38-darwin.so', 'cython_naive_bayes.c', 'cython_decision_tree_functions.pyx', 'cython_tsne.pyx', 'cython_unsupervised_clustering.cpython-38-darwin.so', '.DS_Store', 'cython_decision_tree_functions.so', 'cython_unsupervised_clustering.pyx', 'cython_naive_bayes.so', 'cython_tsne.o', 'cython_knn.so', 'cython_knn.c', 'cython_naive_bayes.pyx', 'cython_ensemble_learning.o', 'cython_decision_tree_functions.o', 'cython_ensemble_learning.pyx', 'cython_naive_bayes.o', 'cython_ensemble_learning.so', 'cython_knn.cpython-38-darwin.so', 'cython_tsne.cpython-38-darwin.so', 'cython_tsne.c', 'cython_naive_bayes.cpython-38-darwin.so', 'cython_knn.pyx', 'cython_knn.o', 'cython_tsne.so', 'cython_ensemble_learning.c', 'cython_decision_tree_functions.c']

setup(
  name = 'sealion',
  packages = setuptools.find_packages(),
  package_data = {'' : non_python_files},
  include_package_data=True,
  version = '3.0.5',
  license='MIT',
  description = 'SeaLion is a simple machine learning and data science library for beginners and ml-engineers alike.',
  author = 'Anish Lakkapragada',
  author_email = 'anish.lakkapragada@gmail.com',
  url = 'https://github.com/anish-lakkapragada/SeaLion',
  download_url = 'https://github.com/anish-lakkapragada/SeaLion/archive/v3.0.5.tar.gz',
  keywords = ['Machine Learning', 'Data Science', 'Python'],
  install_requires=[
          'numpy',
          'joblib',
          'pandas',
          'scipy',
          'tqdm',
          'multiprocess'
      ],
  long_description=open('README.md', 'r').read(),
  python_requires='>=3',
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3'
  ],
)



from cython_python_testing import cython_tsne
tsne = cython_tsne.tSNE(1, 0.0014)
import numpy as np
from sklearn.datasets import make_blobs
X, _ = make_blobs(250)
x_transform = tsne.transform(X)

x_transform = [x_t.tolist() + [0] for x_t in x_transform]

import matplotlib.pyplot as plt
plt.cla()

def plot_points(points):
    import matplotlib.pyplot as plt
    plt.scatter([p[0] for p in points], [p[1] for p in points])

plot_points(x_transform)
plot_points(X)


# Two events E and F are independent if
# the probability of them P(E, F) = product of each one or P(E)P(F)
from DataScienceFromScratch.Statistics import mean
from collections import Counter

'''
->> independence : P(E, F) = P(E)P(F)
->> otherwise they are conditional : P(E | F) = P(E, F)/P(F) #the P(event a | event b) means prob of A
given B
->> P(E, F) = P(E|F)P(F)
->> P(E|F) = P(E), when E and F are independent events (being conditional on F doesn't change P(E) in P(E | F)

Famous example :
- >> if there are two children and one is a girl (G) what's the probability both are girls (B)
->> this is P(B|G) = P(B, G)/P(G) = P(B)/P(G) = 1/2 / 1 = 1/2
->>  P(B, G) = P(B) as the two events are independent
->> P(B, G) = P(B)P(G)/P(G) = P(B) = 1/2 anyways...

Slightly harder example :
->> probability of event B given at least one is a girl (L)
->> P(B | L)  = P(B)/P(L) = 1/3!
->> if you know at least 1 of the children is a girl, then it is 2x more likely there is one boy and one girl (0.5)
than just both (0.25), so only 1 out of 3 times will you meet a family with that 0.25 case

Bayes Theorem :
->> P(H|E) = P(E|H)P(H)/ P(E|H)P(H) + P(E!H)P(!H)
->> this could be for instance that if 99% of the time people are diagnosed correctly and 1% has the disease
->> H = actually +, and E is diagnosed positive
then ->> P(actually positive) = 0.99(0.01)/ (0.99 * 0.01 + 0.01 * 0.99) = 0.5% is correctly diagnosed
->> another way to think about this could be that in a population with a thousand people you would expect
->> 10 of them to have the disease and 9 to test positive. Out of the other 990 that don't have the disease
you would expect then that about 10 of them test posiitve. So 9/(9 + 10) shows the test is only 47% accurate.
'''

#example to prove that :
import enum, random
# Enum is a set of items attached to values
class Kid(enum.Enum) :
    BOY = 0
    GIRL = 1
def random_kid() -> Kid  :
    return random.choice([Kid.BOY, Kid.GIRL])

both_girls = older_girl = either_girl = 0

for _ in range(10000) :
    younger = random_kid()
    older = random_kid()
    if older == Kid.GIRL :
        older_girl += 1
    if older == Kid.GIRL and younger == Kid.GIRL :
        both_girls += 1
    if older == Kid.GIRL or younger == Kid.GIRL :
        either_girl += 1

print("P(both | older) : ", (both_girls/older_girl)) #chance of getting both girls
print("P (both | either) : ", (both_girls/either_girl)) #chance of having both given at least

'''
Note on Random Variables : 
-> implicitly defined in typical data science
->> a random variable is a variable whose possible values have a probability distribution
->> this could be 0 if tails, 1 if heads so var = {0 <- 1/2 and 1 <- 1/2}
->> there is an expected value result, which is just the sum of each prob * value
->> another example of a random variable could be Y where it gives the number of girls conditional on either
->> Y = {1 : 2/3, 2 : 1/3}
->> Z = {1 : 1/2, 2 : 1/2}, Z could be how many boys you have if you already have one
'''

def expected_value(random_variable) :
    '''random_variable is dict'''
    return sum([item * prob for item, prob in random_variable.items()])

assert expected_value({0 : 0.5, 1 : 0.5}) == 0.5
assert expected_value({2 : 0.5, 3 : 0.25, 10 : 0.25}) == 17/4

'''
Notes on PDFs and CDFs : 
->> a probability density function is tellling us the probability of a value leading to a result
->> integral of PDF must equal 1 
->> CDF (cumulative distributon gives the probability) something is less than or equal to a value
->> see page 81, looks like as it goes to 0.5, prob is 0.5 (cumulative from 0.0 - 0.5) and as it goes to 0.9 it's 0.9 (cumulative from 0.0 - 0.9)
'''

def uniform_pdf(x : float) -> float :
    return 1 if 0 <= x < 1 else 0

def uniform_cdf(x : float) -> float :
    '''returns probability that a uniform variable (0 or 1) is less than x'''
    if x < 0 : return 0 #uniform var never less than 0
    elif x < 1 : return x #i.e. if it's 0.5 (cumulative), then prob is 0.5
    else : return 1 #yes because if x >= 1, a uniform var is for sure <= x

'''
Normal Distribution : 
->> shorter skinnier, longer more dull 
->> PDF : f(x | mu, std) = 1/(sqrt(2pi) * std) * exp(-(x-mu)^2/2*std^2))
'''

import math
SQRT_TWO_PI = math.sqrt(2 * math.pi)

def normal_pdf(x: float, mu : float = 0, sigma : float = 1) -> float :
    return (math.exp(-(-x-mu)**2 / 2/ sigma ** 2)/(SQRT_TWO_PI * sigma))

#plot some of these PDFs
import matplotlib.pyplot as plt
xs = [x/10.0 for x in range(-50, 50)]
plt.plot(xs, [normal_pdf(x) for x in xs], '-', label = "mu=0, sigma=1")
plt.plot(xs, [normal_pdf(x, sigma = 2) for x in xs], '--', label = "mu=0, sigma=2")
plt.plot(xs, [normal_pdf(x, sigma = 0.5) for x in xs], ':', label = "mu=0, sigma=0.5")
plt.plot(xs, [normal_pdf(x, mu = -1) for x in xs], '-.', label = "mu=-1, sigma=1")
plt.legend()
plt.title("Various Normal PDFs")
plt.show()

'''notice how shrinking the deviation makes the PDF taller and skinnier'''
'''increasing deviation does the opposite, and changing mean changes the center of the data'''

'''
The standard normal distribution is when mu = 0, std/sigma = 0 (of any PDF - normal or random).
If Z is a standard normal variable, then : 
X = sigmaZ + mu is also normal but with mean u and std sigma (hence why np.random.randn() * sqrt(2/fan_in)) 
creates sqrt(2/fan_in) variance as np.random.random() is already a standard normal distribution

Conversely if X is a normal random variable with mu and sigma : 
Z = (X - mu)/sigma is also standard normal variable
'''

def normal_cdf(x : float, mu : float = 0 , sigma : float =1) -> float :
    '''chance anything selected in the given normal PDF is less than x'''
    return (1 + math.erf((x-mu)/math.sqrt(2)/sigma)) / 2

plt.cla()
plt.plot(xs, [normal_cdf(x, sigma=1) for x in xs], '-', label = 'mu=0, sigma = 1')
plt.plot(xs, [normal_cdf(x, sigma=2) for x in xs], '--', label = 'mu=0, sigma = 2')
plt.plot(xs, [normal_cdf(x, sigma=0.5) for x in xs], ':', label = 'mu=0, sigma = 0.5')
plt.plot(xs, [normal_cdf(x, mu=-1) for x in xs], '-.', label = 'mu=-1, sigma = 1')
plt.legend(loc = 4) #bottom right
plt.title("Various Normal CDFs")
plt.show()

'''
Sometimes we will need to invert normal_cdf to find the value corresponding
to a specified probability. Normal_cdf is continuous and increasing, so you can use 
the binary search algorithm.
'''

def inverse_normal_cdf(p : float, mu :float =0, sigma : float = 1,
                       tolerance : float = 0.000001) -> float :
    '''find approximate inverse using binary search'''
    #if not standard, compute and rescale
    if mu != 0 or sigma != 1 :
        return mu + sigma * inverse_normal_cdf(p, tolerance=tolerance) #cstandardize
    low_z = -10 #normal_cdf(-10) is very close to 0
    high_z = 10 #normal_cdf(10) is very close to 1
    while high_z  - low_z > tolerance :
        mid_z = (low_z + high_z)/2 #find midpoint
        mid_p = normal_cdf(mid_z) #find cdf value there
        if mid_p < p :
            low_z = mid_z #midpoint too low, search above it
        else :
            high_z = mid_z #midpoint too high, search below it
    return mid_z

print(inverse_normal_cdf(0.5))

'''
Why is the normal distribution useful? 
->> sampled averages of any dataset create a normal distribution
'''

xs = [_ for _ in range(-100, 100)]
ys = [2 * x + 1 for x in xs]
sample_size = 30
n_means = 10000
means = []
for _ in range(n_means) :
    start_index = random.randint(0, len(xs) - sample_size)
    sample = ys[start_index : start_index + sample_size]
    means.append(mean(sample))

plt.cla()
frequency_means = Counter(round(mean/10) * 10 for mean in means)
plt.bar(
    [fr + 5 for fr in frequency_means.keys()], frequency_means.values()
)


"""
Why is the normal distribution so useful? 

- the central limit theorem states that any random variable that comes 
from an average of an identically distributed set creates a normal distribution. 

(x1 + ... xN) - mean * n_samples / sigma * sqrt(n_samples) is equal to the normal distribution
where the denominator sigma * sqrt(n_samples) is the std

Binomial Random Variables - > Binomial(n, p) random variable is the sum 
of n bernoulli trials, each with a probability p of 1 and 1 - p of 0 
"""

def bernoulli_trial(p : float) -> int :
    """Returns 1 with probability p and 0 with probability 1 - p"""
    return 1 if random.random() < p else 0 #the smaller the p, the more likely
    # random.random() will be less than p and vice versa.

def binomial(n : int, p : float) -> float :
    '''return the sum of n bernoulli trials with probability p '''
    return sum(bernoulli_trial(p) for trial in range(n))

"""
the mean of a bernoulli(p) variable is p, and its std is sqrt(p(1-p))

the central limit theorem states that as n gets large, a binomial(n, p) 
variable has mu = np, and a standard deviation of sqrt(np(1-p))
"""

#proving this

def binomial_histogram(p : float, n : int, num_points : int) -> None :
    """Pick points from a binomial (n, p) and plot their histogram"""

    data = [binomial(n, p) for _ in range(num_points)]

    #use a bar chart to show the actual binomial samples
    histogram = Counter(data)
    plt.bar([x - 0.4 for x in histogram.keys() ],
            [ v / num_points for v in histogram.values()],
            0.8,
            color = '0.75')

    mu = n * p
    sigma = math.sqrt(n * p * (1 - p)) #todo understand why it is this

    #use a line chart to show the normal approximation
    xs = range(min(data), max(data) + 1)
    ys = [normal_cdf(i + 0.5, mu, sigma) - normal_cdf(i - 0.5, mu , sigma)
          for i in xs] #this is because of continuity correction we have the +/- 0.5
    plt.plot(xs, ys)
    plt.title("binomial distribution vs. normal approximation")
    plt.show()

binomial_histogram(0.5, 100, 10000)



"""
Working with 1D data is the easiest. Instead of finding mean, min, std, etc. it's more interesting to bucketize your data
and count how many points fall into each bucket and put that in a histogram.
"""

from typing import List, Dict
from collections import Counter
import math

import matplotlib.pyplot as plt


def bucketize(point: float, bucket_size: float) -> float:
    """Floor point to the next lower multiple of bucket_size."""
    return bucket_size * math.floor(point / bucket_size)


def make_histogram(points: List[float], bucket_size: float) -> Dict[float, int]:
    """Bucket the points and count how many points are in each bucket"""
    return Counter(bucketize(point, bucket_size) for point in points)  # for each point put it into a bucket and count
    # how many in each bucket


def plot_histogram(points: List[float], bucket_size: float, title: str = ""):
    histogram = make_histogram(points, bucket_size)
    plt.bar(histogram.keys(), histogram.values(), width=bucket_size)
    plt.title(title)


# considering 2 datasets below
import random
from DataScienceFromScratch.Probability import inverse_normal_cdf

random.seed(0)

# uniform between -100 and 100
uniform = [200 * random.random() - 100 for _ in range(10000)]

# normal distribution with mean 0, std 57
normal = [57 * inverse_normal_cdf(random.random())
          # what values give probablities from 0 - 1 in the standard normal CDF (mu = 0, std = 1)
          for _ in range(10000)]  # std * x + mu

plot_histogram(uniform, 10, title="Uniform Histogram")
plt.cla()  # clear
plot_histogram(normal, 10, title="Normal Histogram")
plt.cla()

"""
Having a dataset in 2-Dimensions : (understand each dimension individually)
"""


def random_normal() -> float:
    """Returns a value leading to a random decimal probability in a standard normal distribution"""
    return inverse_normal_cdf(random.random())


xs = [random_normal() for _ in range(1000)]
ys1 = [x + random_normal() / 2 for x in xs]
ys2 = [-x + random_normal() / 2 for x in xs]

# despite both being normally distributed, they have different shapes

plt.scatter(xs, ys1, marker=".", color="red", label="ys1")
plt.scatter(xs, ys2, marker=".", color="green", label="ys2")
plt.xlabel("xs")
plt.ylabel("ys")
plt.legend(loc=9)
plt.title("Very Different Joint Distributions")
plt.show()

# plt.cla()

# looking at correlation to find the difference
from DataScienceFromScratch.Statistics import correlation

print("XS & YS1 correlation : ", correlation(xs, ys1))
print("XS & YS2 correlation : ", correlation(xs, ys2))

# many more dimensions.
"""
If you have many more dims, you would like to know how all dims relate 
to one another. Taking a look at the correlation matrix is a way to see how the ith row 
and jth column.
"""

from DataScienceFromScratch.LinearAlgebra import Matrix, Vector, make_matrix, get_column, shape


def correlation_matrix(data: List[Vector]) -> Matrix:
    """Returns the len(data) * len(data) matrix whose of i-j entry is the correlation
    of the i and j dimensions"""

    def correlation_ij(i: int, j: int) -> float:
        return correlation(get_column(data, i), get_column(data, j))  # correlation between the ith

    return make_matrix(len(data), len(data), correlation_ij)


# making a scatterplot matrix

def create_data(rows: int, cols: int) -> Matrix:
    return [[random.random() for col in range(cols)]
            for row in range(rows)]


corr_data = create_data(4, 4)
_, num_cols = shape(corr_data)
fig, ax = plt.subplots(num_cols, num_cols)
for i in range(num_cols):
    for j in range(num_cols):

        # scatter column_j on x-axis vs. column-i on the y-axis
        if i != j:
            ax[i][j].scatter(corr_data[j], corr_data[i])

        # if they are equal (the diagonal), show the series name
        else:
            ax[i][j].annotate("series " + str(i), (0.5, 0.5),
                              xycoords="axes fraction",
                              ha="center", va="center")

        # hide axis labels except for the left and bottom charts
        if i < num_cols - 1: ax[i][j].xaxis.set_visible(False)
        if j > 0: ax[i][j].yaxis.set_visible(False)

# Fix the bottom-right and top-left axis labels, which are wrong because their charts have only text in them
ax[-1][-1].set_xlim(ax[0][-1].get_xlim())
ax[0][0].set_ylim(ax[0][1].get_ylim())

plt.show()

# Named Tuples

# a common way of representing data is through dicts like :
import datetime

stock_price = {"closing_price": 102.06,
               "date": datetime.date(2021, 1, 1),
               "symbol": "New Year"}

# inefficient for storage, and accessing can go wrong, e.g : stoc_price['cosing_price'] = 102 (typo)

# you don't know what is stored where in a dictionary, e.g. what every key means

prices: Dict[datetime.date, float] = {}  # what if there are multiple different data types

# instead, use a named tuple - which is just a tuple but has named slots
from collections import namedtuple

StockPrice = namedtuple("StockPrice", ['Symbol', 'Date',
                                       'Closing_price'])  # name_of_tuple = namedtuple("name_of_tuple", [all_data_in_strings])
price = StockPrice('MSFT', datetime.date(2021, 1, 1), '106.03')

assert price.Symbol == "MSFT"
assert price.Closing_price == '106.03'  # easy access to anything here

# however, these named tuples are immutable.

# for typing annotations you can use NamedTuple
from typing import NamedTuple


class StockPrice(NamedTuple):
    Symbol: str
    Date: datetime.date
    Closing_price: float

    def is_new_year(self) -> bool:
        return self.Date == datetime.date(2021, 1, 1)


price = StockPrice("MSFT", datetime.date(2021, 1, 1), 200.0)

assert price.Symbol == "MSFT"
assert price.Closing_price == 200
assert price.is_new_year()

# dataclasses, added functions to NamedTuples
from dataclasses import dataclass


@dataclass
class StockPrice2:
    Symbol: str
    Date: datetime.date
    Closing_price: float

    def is_new_year(self) -> bool:
        return self.Date == datetime.date(2021, 1, 1)


price2 = StockPrice2("AAPL", datetime.date(2021, 1, 1), 2000)
assert price2.Symbol == "AAPL"
assert price2.Date == datetime.date(2021, 1, 1)
assert price2.is_new_year()

# we can modify a data class values
price2.Date = datetime.date(2021, 1, 2)
assert not price2.is_new_year()

# unfortunately, this doesn't give an error : price2.cosing_price = 200 (typo allowed)

# cleaning and munging data

from dateutil.parser import parse


def parse_row(row: List[str]) -> StockPrice:
    Symbol, Date, Closing_price = row
    return StockPrice(Symbol=Symbol,
                      Date=parse(Date).date(),  # turns "2020-1-1" -> datetime.date(2020, 1, 1)
                      Closing_price=float(Closing_price))


stock = parse_row(["MSFT", "2020-1-1", "200.54"])
print("This is the stock : ", stock)
assert stock.Symbol == "MSFT"
assert stock.Closing_price == 200.54
assert stock.Date == datetime.date(2020, 1, 1)

# if there's some bad data (e.g. a string in a numeric col or vice versa) - maybe you want to turn that to None and
# avoid the error(s) ?

from typing import Optional
import re


def try_parse_row(row: List[str]) -> Optional[StockPrice]:
    Symbol, Date_, Closing_price_ = row

    # Stock symbol should be all CAPS and only letters
    if not re.match(r"^[A-Z]+$", Symbol):
        return None

    try:
        Date = parse(Date_).date()
    except ValueError:
        return None

    try:
        Closing_price = float(Closing_price_)
    except ValueError:
        return None

    return StockPrice(Symbol, Date, Closing_price)


assert try_parse_row(["msft", '2021-1-1', '200']) is None
assert try_parse_row(["MSFT", '2019-2030-02', 2031]) is None
assert try_parse_row(["MSFT", "2020-31-12", "?"]) is None
assert try_parse_row(['TSLA', '2021-1-1', '200.56']) is not None
assert try_parse_row(["MSFT", "2020-1-1", "200.54"]) == stock

# now we can read and return only the good data

import csv

data: List[StockPrice] = []

with open("comma_delimited_stock_prices.csv") as f:
    reader = csv.reader(f)
    for row in reader:
        maybe_stock = try_parse_row(row)  # may return None or all will be correct
        if not maybe_stock:
            print(f"skipping invalid row : {row}")
        else:
            data.append(maybe_stock)

print("This is the data : ", data)

# Manipulating Data

# easy way to find the highest closing price for AAPL stock

max_aapl_price = max(stock_price.Closing_price for stock_price in data if stock_price.Symbol == "AAPL")

# to find the highest closing price for each stock

from collections import defaultdict

max_prices: Dict[str, float] = defaultdict(lambda: float('-inf'))  # -inf if the price is invalid because anything else
# is greater than that and will take it's place

for sp in data:
    Symbol, Closing_price = sp.Symbol, sp.Closing_price
    if Closing_price > max_prices[Symbol]:
        max_prices[Symbol] = Closing_price

print("Max prices : ", max_prices)

#What is the smallest and biggest 1-day percent changes?

#collect the prices by symbol
prices : Dict[str, List[StockPrice]] = defaultdict(list)

for sp in data :
    prices[sp.Symbol].append(sp)

prices = {symbol : sorted(symbol_prices) for symbol, symbol_prices in prices.items()} #sorts in order, so if there are
# multiple same symbols next it gets sorted by date

print("These are prices : ", prices)

def pct_change(yesterday : StockPrice, today : StockPrice) -> float :
    return today.Closing_price / yesterday.Closing_price - 1 #gives the percent change

class DailyChange(NamedTuple) :
    Symbol : str
    Date : datetime.date
    pct_change : float

def day_over_day_changes(prices : List[StockPrice]) -> List[DailyChange] :
    return [DailyChange(Symbol = today.Symbol,
                        Date = today.Date,
                        pct_change = pct_change(yesterday, today)) #percent change
                    for yesterday, today in zip(prices, prices[1:])] #give a named tuple w/ type annotations of yesterday and today changes

all_changes = [change
               for symbol_prices in prices.values()
               for change in day_over_day_changes(symbol_prices)]

print("This is all_changes : ", all_changes)

#find the largest and smallest changes
max_change = max(all_changes, key = lambda change : change.pct_change) #take the named_tuple and find it's change
print("Single max_change : ", max_change)
assert max_change.Symbol == "FB"

min_change = min(all_changes, key = lambda all_change : all_change.pct_change)
print("Single min change : ", min_change)

changes_by_month : Dict[int, List[DailyChange]] = {month : [] for month in range(1, 13)}

for change in all_changes :
    '''take every Daily Change item and put it into the changes_by_month by its month'''
    changes_by_month[change.Date.month].append(change)

avg_daily_changes = {
    month : sum(change.pct_change for change in changes) / len(changes) if len(changes) else 0  #for each month give the average of its change (changes is a List[DailyChange] item)
    for month, changes in changes_by_month.items()
}

print("Best change : ", max(avg_daily_changes.values()))
#to find the best month
changes_to_month = {changes : month for month, changes in avg_daily_changes.items()}
print("Best month : ", changes_to_month[max(changes_to_month)]) #at the biggest change, what was the best month (looks like there was no change?)

#many techniques require rescaling and here's why :

from DataScienceFromScratch.LinearAlgebra import distance

#measuring distance of points
a_to_b = distance([63, 150], [67, 160])
a_to_c = distance([63, 150], [70, 171])
b_to_c = distance([67, 160], [70, 171])
print("Inches : ")
print(a_to_b)
print(a_to_c)
print(b_to_c)

#measuring distance, now with cm instead of inches
a_to_b = distance([160, 150], [170.2, 160])
a_to_c = distance([160, 150], [177.8, 171])
b_to_c = distance([170.2, 160], [177.8, 171])
print("Centimeters : ")
print(a_to_b)
print(a_to_c)
print(b_to_c)

"""
This a problem, because the units can change distances like that. We may not be able to always convert units, so instead
we can rescale so each dimension has a mean of 0 and an std of 1. This gets rid of the units, and everything is measured
in standard deviations from the mean (much more effective.)
"""

from typing import Tuple
from DataScienceFromScratch.LinearAlgebra import vector_mean
from DataScienceFromScratch.Statistics import standard_deviation

def scale(data : List[Vector]) -> Tuple[Vector, Vector] :
    """returns the mean and std for each position"""
    means = vector_mean(data)
    dims = len(data[0]) #no. of dimensions
    stdev = [standard_deviation(get_column(data, i)) for i in range(dims)] #std of each dimension
    return means, stdev

vectors = [[-3, -1, 1], [-1, 0, 1], [1, 1, 1]]
means, stdevs = scale(vectors)
assert means == [-1, 0, 1]
assert stdevs == [2, 1, 0]


def rescale(data : List[Vector]) -> List[Vector] :
    """Rescales so input data has mu 0 , std 1."""
    dim = len(data[0])
    means, stdevs = scale(data)

    #make a copy of each vector (don't edit previous one)
    rescaled = [v[:] for v in data]

    for v in rescaled :
        for i in range(dim) :
            if stdevs[i] > 0: #if there's no deviation, don't rescale
                v[i] = (v[i] - means[i])/stdevs[i]

    return rescaled


means, stdevs = scale(rescale(vectors))
assert means == [0, 0, 1] #no deviation on 3rd dim
assert stdevs == [1, 1, 0] #no deviation in 3rd dim

#you may have to use your judgemnet, because let's say you're looking at data in a range from 3.5 to 4, the variation could just be noise
# so why make it as important as other dimensions (hide the fact its noise to the machine)

#an aside : using tqdm
import tqdm

for i in tqdm.tqdm(range(100)) :
    _ = [random.random() for _ in range(100000)]

#setting the description bar while it is running
from typing import List

def primes_up_to(n : int) -> List[int] :
    primes = [2]
    with tqdm.trange(3, n) as t :
        for i in t :
            #i is prime if no smaller prime divides it
            i_is_prime = not any(i % p == 0 for p in primes)
            if i_is_prime :
                primes.append(i)
            t.set_description(f"{len(primes)} primes")
    return primes

primes_100 = primes_up_to(100)

#PCA w/o the SVD
"""
Find an axis that captures the most variance as possible.
"""

from DataScienceFromScratch.LinearAlgebra import subtract

def de_mean(data : List[Vector]) -> List[Vector] :
    """recenter data to get mean 0 in each dim"""
    mean = vector_mean(data)
    return [subtract(vector, mean) for vector in data]

from DataScienceFromScratch.LinearAlgebra import magnitude

def direction(w : Vector) -> Vector :
    '''direction is just a magnitude 1 vector, so rescale it here '''
    magnitude_w = magnitude(w)
    return [w_i / magnitude_w for w_i in w]

assert round(magnitude(direction([3, 4, 5]))) == 1

#due to direction, we can calculat ethe variance of data in the direction determined by nonzero w

from DataScienceFromScratch.LinearAlgebra import dot

def directional_variance(data : List[Vector], w : Vector) -> float :
    '''returns variance of x in the direction w'''
    w_dir = direction(w)
    return sum(dot(v, w_dir) ** 2 for v in data) #for v in data find how much v extends in the w direction ** 2

#todo - why no N-1

def directional_variance_gradient(data : List[Vector], w : Vector) -> Vector :
    """gradient of directional variance wrt w"""
    w_dir = direction(w)
    return [sum(2 * dot(v, w_dir) * v[i] for v in data) for i in range(len(w))] #for each dimension see which direction
    #maximizes the variance

from DataScienceFromScratch.GradientDescent import gradient_step

def first_principal_component(data : List[Vector],
                              n : int = 100,
                              step_size : float = 0.1) -> Vector :

    #start out with a random guess
    guess = [1.0 for _ in data[0]]

    with tqdm.trange(n) as t :
        for _ in t :
            dv = directional_variance(data, guess)
            gradient = directional_variance_gradient(data, guess)
            guess = gradient_step(guess, gradient, step_size)
            t.set_description(f" dv : {dv:.3f}")

    return direction(guess)

from DataScienceFromScratch.LinearAlgebra import scalar_multiply

def project(v : Vector, w : Vector) -> Vector :
    """return the projection of v onto w"""
    projection_length = dot(v, w)
    return scalar_multiply(projection_length, w)

from DataScienceFromScratch.LinearAlgebra import subtract

def remove_projection_from_vector(v : Vector, w : Vector) -> Vector :
    """projects  v onto w and subtracts the result from v"""
    return subtract(v, project(v, w))

def remove_projection(data : List[Vector], w : Vector) -> List[Vector] :
    return [remove_projection_from_vector(v, w) for v in data]

#this makes the dataset 1D, but for higher dimensional datasets :

def pca(data : List[Vector], num_components : int) -> List[Vector] :
    components : List[Vector] = []
    for _ in range(num_components) :
        component = first_principal_component(data)
        components.append(component)
        data = remove_projection(data, component)
    return components

#transform the data

def transform_vector(v : Vector, components : List[Vector]) -> Vector :
    return [dot(v, w) for w in components]

def transform(data : List[Vector], components : List[Vector]) -> List[Vector] :
    return [transform_vector(v, components) for v in data]

#do PCA on X


X = [
    [20.9666776351559,-13.1138080189357],
    [22.7719907680008,-19.8890894944696],
    [25.6687103160153,-11.9956004517219],
    [18.0019794950564,-18.1989191165133],
    [21.3967402102156,-10.8893126308196],
    [0.443696899177716,-19.7221132386308],
    [29.9198322142127,-14.0958668502427],
    [19.0805843080126,-13.7888747608312],
    [16.4685063521314,-11.2612927034291],
    [21.4597664701884,-12.4740034586705],
    [3.87655283720532,-17.575162461771],
    [34.5713920556787,-10.705185165378],
    [13.3732115747722,-16.7270274494424],
    [20.7281704141919,-8.81165591556553],
    [24.839851437942,-12.1240962157419],
    [20.3019544741252,-12.8725060780898],
    [21.9021426929599,-17.3225432396452],
    [23.2285885715486,-12.2676568419045],
    [28.5749111681851,-13.2616470619453],
    [29.2957424128701,-14.6299928678996],
    [15.2495527798625,-18.4649714274207],
    [26.5567257400476,-9.19794350561966],
    [30.1934232346361,-12.6272709845971],
    [36.8267446011057,-7.25409849336718],
    [32.157416823084,-10.4729534347553],
    [5.85964365291694,-22.6573731626132],
    [25.7426190674693,-14.8055803854566],
    [16.237602636139,-16.5920595763719],
    [14.7408608850568,-20.0537715298403],
    [6.85907008242544,-18.3965586884781],
    [26.5918329233128,-8.92664811750842],
    [-11.2216019958228,-27.0519081982856],
    [8.93593745011035,-20.8261235122575],
    [24.4481258671796,-18.0324012215159],
    [2.82048515404903,-22.4208457598703],
    [30.8803004755948,-11.455358009593],
    [15.4586738236098,-11.1242825084309],
    [28.5332537090494,-14.7898744423126],
    [40.4830293441052,-2.41946428697183],
    [15.7563759125684,-13.5771266003795],
    [19.3635588851727,-20.6224770470434],
    [13.4212840786467,-19.0238227375766],
    [7.77570680426702,-16.6385739839089],
    [21.4865983854408,-15.290799330002],
    [12.6392705930724,-23.6433305964301],
    [12.4746151388128,-17.9720169566614],
    [23.4572410437998,-14.602080545086],
    [13.6878189833565,-18.9687408182414],
    [15.4077465943441,-14.5352487124086],
    [20.3356581548895,-10.0883159703702],
    [20.7093833689359,-12.6939091236766],
    [11.1032293684441,-14.1383848928755],
    [17.5048321498308,-9.2338593361801],
    [16.3303688220188,-15.1054735529158],
    [26.6929062710726,-13.306030567991],
    [34.4985678099711,-9.86199941278607],
    [39.1374291499406,-10.5621430853401],
    [21.9088956482146,-9.95198845621849],
    [22.2367457578087,-17.2200123442707],
    [10.0032784145577,-19.3557700653426],
    [14.045833906665,-15.871937521131],
    [15.5640911917607,-18.3396956121887],
    [24.4771926581586,-14.8715313479137],
    [26.533415556629,-14.693883922494],
    [12.8722580202544,-21.2750596021509],
    [24.4768291376862,-15.9592080959207],
    [18.2230748567433,-14.6541444069985],
    [4.1902148367447,-20.6144032528762],
    [12.4332594022086,-16.6079789231489],
    [20.5483758651873,-18.8512560786321],
    [17.8180560451358,-12.5451990696752],
    [11.0071081078049,-20.3938092335862],
    [8.30560561422449,-22.9503944138682],
    [33.9857852657284,-4.8371294974382],
    [17.4376502239652,-14.5095976075022],
    [29.0379635148943,-14.8461553663227],
    [29.1344666599319,-7.70862921632672],
    [32.9730697624544,-15.5839178785654],
    [13.4211493998212,-20.150199857584],
    [11.380538260355,-12.8619410359766],
    [28.672631499186,-8.51866271785711],
    [16.4296061111902,-23.3326051279759],
    [25.7168371582585,-13.8899296143829],
    [13.3185154732595,-17.8959160024249],
    [3.60832478605376,-25.4023343597712],
    [39.5445949652652,-11.466377647931],
    [25.1693484426101,-12.2752652925707],
    [25.2884257196471,-7.06710309184533],
    [6.77665715793125,-22.3947299635571],
    [20.1844223778907,-16.0427471125407],
    [25.5506805272535,-9.33856532270204],
    [25.1495682602477,-7.17350567090738],
    [15.6978431006492,-17.5979197162642],
    [37.42780451491,-10.843637288504],
    [22.974620174842,-10.6171162611686],
    [34.6327117468934,-9.26182440487384],
    [34.7042513789061,-6.9630753351114],
    [15.6563953929008,-17.2196961218915],
    [25.2049825789225,-14.1592086208169]
] #all the data

components = pca(X, 1)
X_transform = transform(X, components)

plt.cla()
plt.scatter([x_i[0] for x_i in X], [x_i[1] for x_i in X])
plt.scatter([x_i for x_i in X_transform], [0 for x_i in X_transform])
plt.show()

#PCA's really useful, but a little less interpretable







import numpy as np
from train import train
from nn import NeuralNet
from layers import Linear, Tanh, Sigmoid

inputs = np.array([
    [0, 0],
    [1, 0],
    [0, 1],
    [1, 1]
])

targets = np.array([
    [1, 0],
    [0, 1],
    [0, 1],
    [1, 0]
])

targets = targets[:, 1].reshape(len(targets), 1)

net = NeuralNet([
    Linear(input_size  = 2, output_size = 2),
    Sigmoid(),
    Linear(input_size = 2, output_size = 1),
    Sigmoid()
])

from loss import BinaryCrossentropy, MSE
from optimizers import SGD
train(net, inputs, targets, loss = BinaryCrossentropy(), optimizer = SGD(0.1))

for x, y in zip(inputs, targets) :
    predicted = net.forward(x)
    print(x, predicted, y)


"""
A loss function measures how good our predictions are.
We can use this to adjust the params.
"""

import numpy as np

from tensor import Tensor

class Loss :
    def loss(self, predicted : Tensor, actual : Tensor) -> float :
        raise NotImplementedError
    def grad(self, predicted : Tensor, actual : Tensor) -> Tensor :
        raise NotImplementedError


class MSE(Loss) :
    def loss(self, predicted : Tensor, actual : Tensor) -> float :
        error = np.sum((predicted - actual) ** 2)
        return error
    def grad(self, predicted : Tensor, actual : Tensor) -> Tensor :
        return 2 * (predicted - actual)

class BinaryCrossentropy(Loss) :
    def loss(self, predicted : Tensor, actual : Tensor) -> float :
        error = np.sum((predicted - actual) ** 2)
        return error
    def grad(self, predicted : Tensor, actual : Tensor) -> Tensor :
        return - ((actual/predicted) - (1-actual)/(1-predicted))


from collections import Counter
import numpy as np
cimport numpy as np
import math

cdef list labels
cpdef dict probabilities(list labels) :
    cdef dict count_labels = {label : num_counts/len(labels) for label, num_counts in dict(Counter(labels)).items()}
    return count_labels #{label : num_times}

cpdef double gini_impurity(list labels) :
    cdef dict class_probabilities = probabilities(labels)
    cdef double _, prob
    return 1 - sum([prob ** 2 for _, prob in class_probabilities.items()])

cpdef find_best_split(list data, list labels, int min_samples) :
    """
    :param data: Incoming data
    :return: 2-element tuple (x, y) where x is the index of the feature, and y is the value of that feature for that split.
    We measure how good a split is by how low the new entropy is.
    """

    cdef double incoming_impurity = gini_impurity(labels)
    cdef dict impurity_to_split = {} #impurity : store the (x, y) tuple split]
    cdef dict impurity_to_num_samples = {} #store each impurity to its num_samples
    cdef double label, percent_true, percent_false, impurity_of_split, best_impurity, num_true, num_false
    cdef list input, true_branch, false_branch, observation
    cdef tuple current_split
    cdef int feature
    for input in data :
        for feature in range(len(input)) :
            '''entropy calculated as num_false * false_impurity + num_true * true_impurity'''
            current_split = (feature, input[feature])
            true_branch, false_branch = [], []
            for observation, label in zip(data, labels) :
                if isinstance(observation[current_split[0]], list) :
                    '''this means that it is a one-hot-encoded label, so a data is in the true branch if it is equal to it'''
                    if observation[current_split[0]] == current_split[1] : true_branch.append(label)
                    else : false_branch.append(label)
                elif isinstance(float(observation[current_split[0]]), float) :
                    if observation[current_split[0]] >= current_split[1] : true_branch.append(label)
                    else : false_branch.append(label)
            num_true = len(true_branch)
            num_false = len(false_branch)
            percent_true = num_true/len(data)
            percent_false = num_false/len(data)
            assert percent_false + percent_true == 1, "should be equal to 1"
            impurity_of_split = percent_true * gini_impurity(true_branch) + percent_false * gini_impurity(false_branch)
            impurity_to_split[impurity_of_split] = current_split
            impurity_to_num_samples[impurity_of_split] = len(false_branch) + len(true_branch)

    #find the split ( x, y combo) leading to the least entropy
    best_impurity = min(impurity_to_split)
    if best_impurity < incoming_impurity :
        return impurity_to_split[best_impurity]
    if incoming_impurity >= best_impurity or impurity_to_num_samples[best_impurity] < min_samples :
        '''no improvement, stop and create a leaf node by returning the most common label'''
        return Counter(labels).most_common(1)[0][0]

cpdef tuple branch_data(list data, list labels, tuple split) :
    """
    :param data: inputs
    :param split: (x, y) combination to split data
    :return: true_branches, true_labels, false_branches, false_labels (containing the inputs which were false and true on the given split)
    """
    cdef list true_branch, false_branch, true_labels, false_labels
    true_branch, false_branch = [], []
    true_labels, false_labels = [], []
    cdef list observation
    cdef double label
    for observation, label in zip(data, labels):
        if isinstance(split[1], list) :
            if observation[split[0]] == split[1] :
                true_branch.append(observation)
                true_labels.append(label)
            else :
                false_branch.append(observation)
                false_labels.append(label)
        elif isinstance(float(split[1]), float) :
            if observation[split[0]] >= split[1] :
                true_branch.append(observation)
                true_labels.append(label)
            else :
                false_branch.append(observation)
                false_labels.append(label)
    return true_branch, true_labels, false_branch, false_labels



cpdef double go_down_the_tree(list data_point, dict recursive_tree):
        '''takes in only one data point. Goes down the tree by finding whether it is on the true or false side of each branch and going down
        down that direction recursively until it hits a leaf node'''
        try:
            data_point = np.array(data_point).tolist()
        except Exception:
            pass
        cdef tuple split = list(recursive_tree.keys())[0]  # get the current spit
        cdef bint true_or_false
        try:
            if isinstance(float(split[1]), float):
                true_or_false = data_point[split[0]] >= split[1]
        except Exception:
            if isinstance(anti_tupelize(split[1]), tuple):
                true_or_false = list(data_point[split[0]]) == list(split[1])
        subtree = recursive_tree[split][true_or_false]
        try:
            if isinstance(float(subtree), float):
                return subtree  # its a leaf node
        except Exception:
            return go_down_the_tree(data_point, subtree)


cpdef np.ndarray[np.float64_t, ndim=1] chunk_predict(list data_chunk_with_tree):
    cdef dict tree
    cdef list data_chunk
    data_chunk, tree = data_chunk_with_tree

    cdef list predictions = [go_down_the_tree(data_point, tree) for data_point in data_chunk]
    return np.array(predictions)


cpdef tuple tuplelize(tpl) :
    """Change tpl (1, [2, 3]) to (1, (2, 3)) so it can be stored in a dictionary (the tree dict)"""
    cdef list new_tuple = []
    for element in tpl :
        if isinstance(element, list) :
            new_tuple.append(tuple(element))
        else : new_tuple.append(element)
    return tuple(new_tuple)

cpdef tuple anti_tupelize(tuple tpl) :
    """Take in tuple (1, (2, 3)) for instance in the tree and make it (1, [2, 3]) as the y (which is the value) is a list, not tuple in given dataset"""
    cdef list new_tuple = []
    for element in tpl :
        if isinstance(element, tuple) :
            new_tuple.append(list(element))
        else : new_tuple.append(element)
    return tuple(new_tuple)


cdef class CythonTrainDecisionTraining() :
    cdef dict splits_to_most_common_labels
    cdef dict data_points_to_count
    def __cinit__(self) :
        self.splits_to_most_common_labels = {}
        self.data_points_to_count = {}

    cpdef build_tree(self, list data, list labels, int min_samples, double max_branches = math.inf):
        '''build tree recursively inside this method'''
        try :
            data = np.array(data).tolist()
            labels = np.array(labels).tolist()
        except Exception :
            pass
        best_split = find_best_split(data, labels, min_samples)
        cdef list true_branch, true_labels, false_branch, false_labels
        cdef double most_common_label
        try:
            x = best_split/2
            return best_split
        except Exception:
            true_branch, true_labels, false_branch, false_labels = branch_data(data, labels, best_split)
            self.splits_to_most_common_labels[tuplelize(best_split)] = \
            Counter(true_labels + false_labels).most_common(1)[0][0]
            for point in true_branch :
                if tuplelize(point) in self.data_points_to_count :
                    self.data_points_to_count[tuplelize(point)] += 1
                    if self.data_points_to_count[tuplelize(point)] >= max_branches :
                        most_common_label = Counter(list(labels)).most_common(1)[0][0]
                        return most_common_label
                else :
                    self.data_points_to_count[tuplelize(point)] = 1
            tree = {tuplelize(best_split): {True: CythonTrainDecisionTraining.build_tree(self, true_branch, true_labels, min_samples,  max_branches),
                                            False: CythonTrainDecisionTraining.build_tree(self, false_branch, false_labels, min_samples, max_branches)}}
            return tree

    cpdef dict get_data_points_to_count(self) :
        return dict(self.data_points_to_count)

    cpdef dict get_splits_to_MCLs(self) :
        return self.splits_to_most_common_labels




"""check egrep.py for information on stdin/stdout"""

#"r" reads in
file_for_reading = open("GettingDataFiles/reading_file.txt", "r")
file_for_reading2 = open("GettingDataFiles/reading_file.txt")

#"w" writes -> destroys any current file with given name
file_for_writing = open("GettingDataFiles/writing_file.txt", "w")

#'a' is append -> adds to the end of the file
file_for_appending = open("GettingDataFiles/appending_file.txt", 'a')

#close files once done
file_for_writing.close()

#always use with a with statement, to make sure you don't forget to close it :
"""
with open(filename) as f : 
    data = function(f)
"""

#another example below
import re
starts_with_hash = 0
with open("GettingDataFiles/hashes_file.txt") as f :
    for line in f : #iterate through each line
        if re.match("^#", line) : #contains a hash ?
            starts_with_hash += 1

print("Starts_with_hash : ", starts_with_hash)

#getting the domain of an email adress
def get_domain(email_adress : str) -> str :
    '''split on @ and return last piece'''
    return email_adress.lower().split("@")[-1]

#test it out
assert get_domain("anish.lakkapragada@gmail.com") == "gmail.com"
assert get_domain("anish.lakkapragada@stanford.edu") == "stanford.edu"

from collections import Counter

with open("GettingDataFiles/email_addresses.txt") as f :
    domain_counts = Counter(get_domain(line.strip()) for line in f if "@" in line)


print(domain_counts)

"""
A word on CSV files : 
import csv
with open("delimited_symbol_stock_prices.txt") as f : 
    tab_reader = csv.reader(f, delimiter = "\t")
    for row in tab_reader : 
        data = row[0]
        symbol = row[1]
        closing_price = float(row[2])
        process(date, symbol, closing_price)

To process this kind of dataset : 
date:ticker:closing_price
6/20/2014:AAPL:90.91
6/20/2014:MSFT:41.68
6/20/2014:FB:64.5

with open('colon_delimited_stock_prices.txt') as f : 
    colon_reader = csv.DictReader(f, delimiter = ":")
    for dict_row in colon_reader : 
        date = dict_row["row"]
        ticker = dict_row["ticker"]
        cp = dict_row["closing_price"]
        #data function 
        
WRITING CSV FILES 

todays_prices = {'AAPL' : 90.91, 'MSFT' : 41.68, 'FB' : 64.5}
with open("comma_delimited_stock_prices", 'w') as f : 
    csv_writer = csv.writer(f, delimiter = ",")
    for stock, price in todays_prices.items() : 
        csv_writer.writerow([stock, price])
"""

#scraping the web!!!

from bs4 import BeautifulSoup
import requests

url = "https://raw.githubusercontent.com/joelgrus/data/master/getting-data.html"

html = requests.get(url).text
soup = BeautifulSoup(html, "html5lib")

#finding the first paragraph with <p> tag
first_paragraph = soup.find("p")

first_paragraph_text = soup.p.text
first_paragraph_words = soup.p.text.split()

#extract a tag's attributes
first_paragraph_id = soup.p["id"]
first_paragraph_id2 = soup.p.get("id") #returns None if no id

all_paragraphs = soup.find_all("p")
paragraphs_with_ids = [p for p in soup("p") if p.get('id')]

important_paragraphs = soup("p", {"class" : "important"})
important_paragraphs2 = soup('p', 'important')
important_paragraphs3 = [p for p in soup('p')
                         if 'important' in p.get('class', [])]

spans_inside_divs = [span
                     for div in soup('div')
                     for span in div('span')] #for each <div> on the page


url = "https://www.house.gov/representatives"
text = requests.get(url).text
soup = BeautifulSoup(text, "html5lib")

all_urls = [a["href"]
            for a in soup('a')
            if a.has_attr('href')]

print("Length URLS : ", len(all_urls))

import re
regex = r"^https?://.*\.house\.gov/?$" #must start with "http://" or "https://" and end in "house.gov" or ".house.gov/"

assert re.match(regex, "http://joel.house.gov")
assert re.match(regex, "https://anish.house.gov")
assert re.match(regex, "http://joel.house.gov/")
assert re.match(regex, "https://anish.house.gov/")
assert not re.match(regex, "https://ankit.house.com/")
assert not re.match(regex, "ankit.house.gov/")
assert not re.match(regex, "https://ankit.house.gov/cornell/")

good_urls = [url for url in all_urls if re.match(regex, url)]

print("Length URLS : ", len(good_urls))

good_urls = list(set(good_urls))

print("Length URLS : ", len(good_urls))

html = requests.get("https://jayapal.house.gov").text
soup = BeautifulSoup(html, "html5lib")

links = {a["href"] for a in soup('a') if 'press releases' in a.text.lower()} #use a set to avoid duplicates

print(links)

from typing import Dict, Set
press_releases : Dict[str, Set[str]] = {}

for house_url in good_urls :
    html = requests.get(house_url).text
    soup = BeautifulSoup(html, "html5lib")
    pr_links = {a["href"] for a in soup('a') if 'press releases' in a.text.lower()}  # use a set to avoid duplicates

    print(f"{house_url} : {pr_links}")
    press_releases[house_url] = pr_links


def paragraph_mentions(text : str, keyword : str) -> bool :
    """
    Returns True if a <p> inside the text mentions {keyword}
    """

    soup = BeautifulSoup(text, "html5lib")
    paragraphs = [p.get_text() for p in soup("p")] #get all paragraphs

    return any(keyword.lower() in paragraph.lower()
               for paragraph in paragraphs) #do any of the words in the paragraph have the keyword?

text = """<body><h1>Facebook</h1><p>Twitter</p>"""
assert paragraph_mentions(text, "twitter") #is inside a paragraph (<p>)
assert not paragraph_mentions(text, "facebook") #not inside a <p>
assert not paragraph_mentions(text, "Facebook")

data_mentions = 0
for house_url, pr_links in press_releases.items() :
    for pr_link in pr_links :
        url = f"{house_url}/{pr_link}" #PR link
        text = requests.get(url).text

        if paragraph_mentions(text, "data") :
            print(f"{house_url}")
            data_mentions += 1
            break

print("Data mentions : ", data_mentions)

#using JSON (Javascript Object Notation)
import json
serialized = """{"title" : "Data Science Book", 
              "author" : "Joel Grus", 
              "publicationYear" : 2019, 
              "topics" : ["data", "science", "data science"]}"""

#parse the JSON to create a Python dict
deserialized = json.loads(serialized)
print(deserialized)
assert deserialized["publicationYear"] == 2019
assert "data science" in deserialized["topics"]

import requests, json
github_user = "joelgrus"
endpoint = f"https://api.github.com/users/{github_user}/repos"
repos = json.loads(requests.get(endpoint).text)

from dateutil.parser import parse
dates = [parse(repo["created_at"]) for repo in repos]
month_counts = Counter(date.month for date in dates)
weekday_counts = Counter(date.weekday() for date in dates)

last_5_repos = sorted(repos,
                      key = lambda r : r["pushed_at"],
                      reverse = True)[:5]

last_5_langs = [repo["language"]
                for repo in last_5_repos]

CONSUMER_KEY = "rC6SdpUepwfLNmrd2k1kgTAan"
CONSUMER_SECRET = "cjHW1eKIZJDx5WrirgLGdcPX4yykQwUFmBLRf0fZodmLSM6ipi"

import os, webbrowser
from twython import Twython

#get a temporary client to retrieve an authentication URL
temp_client = Twython(CONSUMER_KEY, CONSUMER_SECRET)
temp_creds = temp_client.get_authentication_tokens()
url = temp_creds['auth_url']

#visit the frickin url to authorize the application to get a pin
print(f"go visit {url} and get the PIN code and paste it below")
webbrowser.open(url)
PIN_CODE = input("please enter the PIN code : ")

#now use the PIN_CODE to get the actual tokens
auth_client = Twython(CONSUMER_KEY,
                      CONSUMER_SECRET,
                      temp_creds["oauth_token"],
                      temp_creds["oauth_token_secret"])
final_step = auth_client.get_authorized_tokens(PIN_CODE)
ACCESS_TOKEN = final_step['oauth_token']
ACCESS_TOKEN_SECRET = final_step["oauth_token_secret"]

#get a new twython instance using them now
twitter = Twython(CONSUMER_KEY,
                  CONSUMER_SECRET,
                  ACCESS_TOKEN,
                  ACCESS_TOKEN_SECRET)

#done with twitter authentication
#search for all tweets containing the phrase "data science"
for status in twitter.search(q = "data science")['statuses'] :
    user = status['user']['screen_name']
    text = status['text']
    print(f"{user}:{text}\n")

"""
Notes on the Streaming API : 

-> the twitter search api only shows a few texts, and we want A LOT
-> use the TwythonStreamer class to get many tweets
"""

from twython import TwythonStreamer

tweets = [] #never append data to global vars in practice
max_tweets = 300
class MyStreamer(TwythonStreamer) :
    def on_success(self, data):
        """
        What do we do when Twitter sends us data?
        Here data will be a python dict to represent a single tweet.
        """

        #collect only english tweets
        if data.get('lang') == "en" :
            tweets.append(data)
            print(f"received tweet # {len(tweets)}")

        #stop when we have collected enough
        if len(tweets) >= max_tweets :
            self.disconnect()

    def on_error(self, status_code, data):
        print(status_code, data)
        self.disconnect()


stream = MyStreamer(CONSUMER_KEY, CONSUMER_SECRET, ACCESS_TOKEN, ACCESS_TOKEN_SECRET) #create a new instance

#start consuming public statuses that have the keyword "data"
stream.statuses.filter(track = "data")

#or get a sample of all tweets
stream.statuses.sample()

from collections import Counter
top_hashtags = Counter(hashtag["text"].lower()
                       for tweet in tweets
                       for hashtag in tweet['entities']['hashtags'])

print(top_hashtags.most_common(500))




